Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Omega.

Set Implicit Arguments.
Unset Strict Implicit.


(**************************************************************)
(** p, Pred    *)
Definition Pred := State -> Prop.

(** m, MPred    *)
Definition MPred := Mem -> Prop.

(** g, Guarantee    *)
Definition Grt := State -> State -> Prop.


Notation STag := nat    (only parsing).

(** SCAP Spec    *)
Inductive Spec : Type :=
  Ast : Pred -> Grt -> STag -> Spec.

(*** Figure 8, Definitions  *)
(** True *)
Definition mTrue : MPred :=
  fun (M : Mem) => True.

(** emp *)
Definition Aemp : MPred :=
  fun (M : Mem) => M = MemMod.emp.

Definition mEq (M : Mem) : MPred :=
  fun (M0 : Mem) => M = M0.

Definition bang (P : Prop) : MPred :=
  fun (M : Mem) => Aemp M /\ P.

(** l |-> n *)
Definition pto (l : Label) (n : int) : MPred :=
  fun (M : Mem) => MemMod.sgltonM M l n.

Notation "A |-> B" := (pto A B) (left associativity, at level 65).

(** l |-> _ *)
Definition ptox (l : Label) : MPred :=
  fun (M : Mem) => exists n : int, pto l n M.

Notation "A |->_" := (ptox A) (left associativity, at level 65).

(** m1*m2 *)
Definition mstar (m1 m2 : MPred) : MPred :=
  fun (M : Mem) =>
    exists M1 : Mem, exists M2 : Mem,
      MemMod.disjUnion M1 M2 M /\ m1 M1 /\ m2 M2.

Notation "A *^* B" := (mstar A B) (left associativity, at level 68).

Definition ptoW (l : Label) (n : int) : MPred :=
  mstar (pto l n) (ptox (l+1)).

Notation "l |->w n" := (ptoW l n) (left associativity, at level 65).

Definition ptoxW (l : Label) : MPred :=
  mstar (ptox l) (ptox (l+1)).

Notation "l |->w_" := (ptoxW l) (left associativity, at level 65).
 
(** l |-> _, ..., _ *)
Fixpoint ptoxBlk (l : Label) (n : nat) {struct n}: MPred :=
  match n with
    0 => Aemp
  | S n' => mstar (ptox l) (ptoxBlk (l+1) n')
  end.

Definition imprecM (m : MPred) : MPred := mstar m mTrue.

(* l ~> n , equivalent to (l -> n) * mTrue *)
Definition leadsTo (l : Label) (n : int) : MPred := imprecM (pto l n).

Notation "l |~> n" := (leadsTo l n) (left associativity, at level 65).

Definition leadsTox (l : Label) : MPred := imprecM (ptox l).

Notation "l |~>_ " := (leadsTox l) (left associativity, at level 65).

Definition leadsToW (l : Label) (n : int) : MPred := imprecM (ptoW l n).

Notation "l |~>w n" := (leadsToW l n) (left associativity, at level 65).

Definition leadsToxW (l : Label) : MPred := imprecM (ptoxW l).

Notation "l |~>w_ " := (leadsToxW l) (left associativity, at level 65).

Definition leadsToBlk (l : Label) (n : nat) : MPred := imprecM (ptoxBlk l n).

Definition stkSpace (sp : int) (n : nat) : MPred :=
  fun (M : Mem) => (sp >= n) /\ (leadsToBlk (sp-n) n) M.

(* precise version of stkSpace *)
Definition stkSpacePrec (sp : int) (n : nat) : MPred :=
  fun (M : Mem) => (sp >= n) /\ (ptoxBlk (sp-n) n) M.


Definition MemUpd16 (M : Mem) (l : int) (n : int) (M' : Mem) : Prop :=
  exists x, M' = MemMod.put (MemMod.put M l n) (l+1) x.

Definition MemUpdW (M : Mem) (l : Label) (w w': nat) : Mem :=
  MemMod.put (MemMod.put M l w) (l+1) w'.

Lemma MemUpd16_MemUpdW : forall M M' l w,
  MemUpd16 M l w M'
  -> exists w', M' = MemUpdW M l w w'.
Proof.
  intros M M' l w H.
  unfold MemUpd16, MemUpdW in * .
  destruct H as [x H].
  subst M'.
  exists x; trivial.
Qed.
Implicit Arguments MemUpd16_MemUpdW [M M' l w].

Lemma MemUpdW_MemUpd16 : forall M M' l w w',
  M' = MemUpdW M l w w'
    -> MemUpd16 M l w M'.
Proof.
  intros M M' l w w' H.
  subst M'.
  unfold MemUpd16, MemUpdW.
  exists w'; trivial.
Qed.
Implicit Arguments MemUpdW_MemUpd16 [M M' l w w'].

(** precise(m)  *)
Definition precise : MPred -> Prop :=
  fun (m : MPred) =>
    forall (M1 M2 M : Mem),
      MemMod.subseteq M1 M -> MemMod.subseteq M2 M -> m M1 -> m M2 -> M1 = M2.

(** p*m *)
Definition pmstar (p : Pred) (m : MPred) : Pred :=
  fun (S : State) =>
    exists M1, exists M2, MemMod.disjUnion M1 M2 (S_Mem S) /\ p (S_Mem_Upd S M1) /\ m M2.

Notation "p *<* m" := (pmstar p m) (left associativity, at level 68).

(** m-*p   *)
Definition pm_mwand (m : MPred) (p : Pred) : Pred :=
  fun (S : State) =>
    forall (M M' : Mem), MemMod.disjUnion M (S_Mem S) M' -> m M -> p (S_Mem_Upd S M').

Notation "m -* p" := (pm_mwand m p) (right associativity, at level 70).

(** p*True imprec *)
Definition imprec (p : Pred) : Pred := pmstar p mTrue.

Definition pTrue : Pred :=
  fun (S : State) => True.


(* the old definition
Definition prepost (m m' : MPred) : Grt :=
  fun (S S' : State) =>
    forall M1 M2, MemMod.disjUnion M1 M2 (S_Mem S) -> m M1 
      -> exists M1', MemMod.disjUnion M1' M2 (S_Mem S') /\ m' M1'.
*)

Definition prepost (m m' : MPred) : Grt :=
  fun (S S' : State) =>
    exists M1, exists M2, exists M1',
      m M1 /\ m' M1' /\ MemMod.disjUnion M1 M2 (S_Mem S) /\ MemMod.disjUnion M1' M2 (S_Mem S').

Definition pimp (p p' : Pred) : Prop :=
  forall (S : State), p S -> p' S.

Definition gimp (g g' : Grt) : Prop :=
  forall (S S' : State), g S S' -> g' S S'.

(*
(** (p, g) => (p', g') **)
Definition simp (s s' : Spec) : Prop :=
  match s with
    Ast p g =>
      match s' with
        Ast p' g' => pimp p p' /\ gimp g' g
      end
  end.
*)

(** Strongest Predicate **)
Definition strongPred (gc: Grt) (p : Pred) : Pred :=
  fun (S : State) =>
  (exists S0, (imprec p) S0 /\ gc S0 S).

(** Weakest Guarantee **)
Definition weakGrt (gc : Grt) (p : Pred) (g : Grt) : Grt :=
  fun (S S' : State) =>
  forall S0, (imprec p) S0 -> gc S0 S -> g S0 S'.

(*
(** Weakest Guarantee **)
Definition weakGrta (gc : Grt) (a : Spec) : Grt :=
  match a with
    Ast p g => weakGrt gc p g
  end.


(** Strongest Assertion **)
Definition strongAst (gc : Grt) (a : Spec) : Spec :=
  match a with
    Ast p g => Ast (strongPred gc p) (weakGrt gc p g)
  end.
*)


Definition enable (p : Pred) (g : Grt) : Prop :=
  forall S, p S -> exists S', g S S'.

Definition conjG (g g' : Grt) : Grt :=
  fun (S S' : State) =>
    exists S'', g S S'' /\ g' S'' S'.

Infix "oo" := conjG (left associativity, at level 69).

Definition linkPG (p : Pred) (g: Grt) : Pred :=
  fun (S' : State) =>
    exists S, p S /\ g S S'.

Infix "|>" := linkPG (left associativity, at level 69). 

Definition conjPG (p : Pred) (g : Grt) : Grt :=
  fun (S S' : State) => p S /\ g S S'.

Infix ">o" := conjPG (left associativity, at level 69).


(** pattern used to define g based on pre and post *)
Definition leadstoPQ (p q : Pred) : Grt :=
  fun (S S' : State) =>
  match S with
    (M, R, fl, isr) =>
      match S' with
        (M', R', fl', isr') =>
            forall M1 M2,
              MemMod.disjUnion M1 M2 M
              -> p (M1, R, fl, isr)
              -> exists M1', MemMod.disjUnion M1' M2 M'
                             /\ q (M1', R', fl', isr')
      end
  end.


Definition enableS (S : State) (g : Grt) : Prop :=
  exists S', g S S'.

(** frame property of g *)
Definition frame (g : Grt) : Prop :=
  forall S S',
    g S S' 
    -> forall M1 M2, MemMod.disjUnion M1 M2 (S_Mem S)
                     -> enableS (S_Mem_Upd S M1) g
                     -> exists M1', MemMod.disjUnion M1' M2 (S_Mem S')
                                    /\ g (S_Mem_Upd S M1) (S_Mem_Upd S' M1').

(** monotonicigy of g *)
Definition safe_mono (g : Grt) : Prop :=
  forall p, enable p g -> enable (fun S => imprec p S) g.
  

Lemma frame_rule_1:
  forall p g p',
    pimp (p |> g) p' -> frame g -> enable p g -> pimp ((fun S=> imprec p S) |> g) (fun S => imprec p' S).
Proof.
  intros.
  unfold pimp. unfold linkPG.
  unfold imprec.
  intros.
  InvertAll. rename x into S0.
  induction S0 as (((M0, R0), fl0), isr0).
  induction S as (((M, R), fl), isr).
  unfold pmstar in H2.
  InvertAll. rename x into M01. rename x0 into M02.
  unfold frame in H0.
  assert (exists M1' : MemMod.Map,
           MemMod.disjUnion M1' M02 (S_Mem (M, R, fl, isr)) /\
           g (S_Mem_Upd (M0, R0, fl0, isr0) M01) (S_Mem_Upd (M, R, fl, isr) M1')).
  apply H0; auto.
  unfold enableS.
  unfold enable in H1. apply H1; auto.
  InvertAll. rename x into M1.
  unfold pmstar. exists M1.  exists M02.
  split; auto. split; auto. UnfoldStateAll. clear H0.
  unfold pimp in H. apply H.
  unfold linkPG. exists (M01, R0, fl0, isr0).
  split; auto.
Qed.

Lemma leadstoPQ_sameQ_reverseP:
  forall p p' q g,
    frame g -> enable p g -> pimp (p |> g) p' 
    -> gimp (g oo (leadstoPQ p' q)) (leadstoPQ p q).
Proof.
  intros.
  unfold gimp.
  intros.
  unfold conjG in H2.
  InvertAll. rename x into S0.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold leadstoPQ.
  induction S0 as (((M0, R0), fl0), isr0).
  intros.
  unfold leadstoPQ in H4.
  unfold frame in H.
  assert (exists M1' : MemMod.Map,
            MemMod.disjUnion M1' M2 (S_Mem (M0, R0, fl0, isr0)) /\
            g (S_Mem_Upd (M, R, fl, isr) M1) (S_Mem_Upd (M0, R0, fl0, isr0) M1')).
  apply H; auto.
  unfold enableS.
  unfold enable in H0. UnfoldStateAll.
  apply H0; auto.
  InvertAll. rename x into M01.
  apply H4 with M01; auto.
  unfold pimp in H1.
  apply H1.
  unfold linkPG. exists (M1, R, fl, isr). split; auto.
Qed.



(*
Inductive SpecX : Type :=
  AstX : Spec -> STag -> SpecX.


Definition SpecX_Tag (thetaX : SpecX) : STag :=
  match thetaX with
    AstX _ t => t
  end.
*)

Lemma imprec_weaker:
  forall (S : State) (p : Pred),
    p S -> imprec p S.
Proof.
  intros.
  unfold imprec.
  unfold pmstar.
  induction S.
  induction a.
  induction a.
  exists a. exists MemMod.emp.
  UnfoldStateAll.
  split. auto with PMap.
  split; auto.
  unfold mTrue. auto.
Qed.
Hint Resolve imprec_weaker.

(* ********** *)
Lemma imprecPimp:
  forall (S : State) (p p0: Pred),
    pimp p p0 -> imprec p S -> imprec p0 S.
Proof.
intros.
unfold pimp in H.
inversion_clear H0.
InvertAll.
unfold imprec.
unfold pmstar.
exists x; exists x0.
split; auto.
Qed.
Hint Resolve imprecPimp.

(*
(** Lemma 2.3 ***)
Lemma StrongAst_Strengthen:
  forall (s s': Spec) (g:Grt),
    simp s s'
    -> simp (strongAst g s) (strongAst g s').
Proof.
intros.
unfold simp in |- *.
unfold strongAst in |- *.
induction s.
induction s'.
inversion_clear H.
split.
 unfold pimp in |- *.
   unfold strongPred in |- *.
   intros.
   inversion_clear H.
   inversion_clear H2.
   rename x into S0; exists S0.
   split.
  apply imprecPimp with p; auto.
  auto.
 unfold gimp, weakGrt in |- *.
   intros.
   unfold gimp in H1.
   apply H1.
   apply H; auto.
   eauto.
Qed.
Hint Resolve StrongAst_Strengthen.
*)

(*** SIMP/PIMP/GIMP Related  Lemmas ***)
Lemma equ_pimp:
  forall (p : Pred), pimp p p.
Proof.
unfold pimp.
intros.
auto.
Qed.

Hint Resolve equ_pimp.

Lemma equ_gimp:
  forall (g : Grt), gimp g g.
Proof.
unfold gimp.
intros.
auto.
Qed.

Hint Resolve equ_gimp.

(*
Lemma equ_simp:
  forall (a : Spec), simp a a.
Proof.
unfold simp.
intros.
induction a.
split; auto.
Qed.

Hint Resolve equ_simp.
*)

Lemma pimp_transt:
  forall (p p' p'' : Pred),   pimp p p' -> pimp p' p'' -> pimp p p''.
Proof.
unfold pimp.
intros.
apply H0.
apply H.
auto.
Qed.

Hint Resolve pimp_transt.


Lemma gimp_transt:
  forall (g g' g'' : Grt),   gimp g g' -> gimp g' g'' -> gimp g g''.
Proof.
unfold gimp.
intros.
apply H0.
apply H.
auto.
Qed.

Hint Resolve gimp_transt.

(*
Lemma simp_transt:
  forall (a a' a'' : Spec), simp a a' -> simp a' a'' -> simp a a''.
Proof.
intros.

unfold simp.
induction a.
induction a'.
induction a''.
inversion_clear H.
inversion_clear H0.
split.
 apply pimp_transt with p0; auto.
 apply gimp_transt with g0; auto.
Qed.

Hint Resolve simp_transt.
*)


Lemma mstar_presv_prec:
  forall (m m' : MPred), precise m -> precise m' -> precise (mstar m m').
Proof.
  intros.
  unfold precise.
  intros.
  unfold mstar in H3.
  unfold mstar in H4.
  InvertAll.
  assert (MemMod.subseteq x M2).
  apply MemMod.disjUnion_subseteq_left with x0; auto.
  assert (MemMod.subseteq x1 M1).
  apply MemMod.disjUnion_subseteq_left with x2; auto.
  assert (MemMod.subseteq x0 M2).
  apply MemMod.disjUnion_subseteq_right with x; auto.
  assert (MemMod.subseteq x2 M1).
  apply MemMod.disjUnion_subseteq_right with x1; auto.
  assert (MemMod.subseteq x1 M).
  apply MemMod.subseteq_transitivity with M1; auto.
  assert (MemMod.subseteq x0 M).
  apply MemMod.subseteq_transitivity with M2; auto.
  assert (MemMod.subseteq x2 M).
  apply MemMod.subseteq_transitivity with M1; auto.
  assert (MemMod.subseteq x M).
  apply MemMod.subseteq_transitivity with M2; auto.
  unfold precise in H. unfold precise in H0.
  assert (x = x1). apply H with M; auto.
  assert (x0 = x2). apply H0 with M; auto.
  unfold MemMod.disjUnion in H5.
  unfold MemMod.disjUnion in H6.
  inversion_clear H5. inversion_clear H6.
  rewrite H17 in H20.
  rewrite H18 in H20.
  rewrite <- H20 in H21. auto.
Qed.
Implicit Arguments mstar_presv_prec [m m'].

Hint Resolve mstar_presv_prec.


Lemma precise_emp : precise Aemp.
Proof.
  unfold Aemp.
  unfold precise.
  intros.
  rewrite <- H1 in H2. auto.
Qed.

Hint Resolve precise_emp.

Lemma mstar_Aemp_unit:
  forall (m: MPred) (M: Mem), mstar m Aemp M -> m M.
Proof.
  intros.
  unfold mstar in H.
  InvertAll.
  unfold Aemp in H2.
  rewrite H2 in H0.
  assert (x = M). auto with PMap.
  rewrite H1 in H. auto.
Qed.

Lemma mstar_Aemp_unit_right:
  forall (m: MPred) (M: Mem), m M -> mstar m Aemp M.
Proof.
  intros.
  unfold mstar.
  exists M. exists MemMod.emp.
  split; auto.
  auto with PMap.
  unfold Aemp. auto.
Qed.
Hint Resolve mstar_Aemp_unit_right.

Lemma pmstar_Aemp_unit:
  forall p S, pmstar p Aemp S -> p S.
Proof.
  intros.
  induction S. induction a. induction a.
  unfold pmstar in H. InvertAll.
  unfold Aemp in H2.
  rewrite H2 in H0.
  UnfoldStateAll.
  assert (x = a). auto with PMap.
  rewrite <- H1. auto.
Qed.

Lemma pmstar_Aemp_unit_right:
  forall (p: Pred) (S: State), p S -> (pmstar p Aemp) S.
Proof.
  intros.
  unfold pmstar.
  induction S. induction a. induction a.
  exists a. exists MemMod.emp. unfold Aemp.
  UnfoldStateAll.
  split; auto with PMap.
Qed.
Hint Resolve pmstar_Aemp_unit_right.

Lemma imprec_monotone:
  forall (p : Pred) (M M': Mem) (R : RFile) (fl : FlagR) (is : ISR),
    imprec p (M, R, fl, is) -> MemMod.subseteq M M' -> imprec p (M', R, fl, is).
Proof.
  intros.
  assert (exists M'', MemMod.disjUnion M M'' M').
  auto with PMap.
  inversion_clear H1.
  unfold imprec.
  unfold pmstar.
  unfold imprec in H.
  unfold pmstar in H.
  UnfoldStateAll.
  InvertAll.
  exists x0.
  exists (MemMod.merge x1 x).
  split.
  unfold MemMod.disjUnion in H1.
  inversion_clear H1.
  rewrite H5 in H2.
  apply MemMod.disjUnion_assoc_left_right; auto.
  split; auto.
Qed.
Hint Resolve imprec_monotone.

Lemma mstar_mwand_equ_left:
  forall m m' p S, pm_mwand (mstar m m') p S -> pm_mwand m (pm_mwand m' p) S.
Proof.
  intros.
  unfold pm_mwand.
  unfold pm_mwand in H.
  intros.
  induction S. induction a. induction a.
  rename a into MS. rename b1 into R. rename b0 into ie. rename b into is.
  UnfoldStateAll.
  assert (MemMod.disjoint M M0).
  unfold MemMod.disjUnion in H0. unfold MemMod.disjUnion in H2. InvertAll.
  rewrite H6 in H4.
  eauto with PMap.
  apply H with (MemMod.merge M M0).
  unfold MemMod.disjUnion in H0. inversion_clear H0.
  rewrite H6 in H2.
  assert (MemMod.merge M MS = MemMod.merge MS M); auto with PMap.
  rewrite H0 in H2.
  auto with PMap.
  unfold mstar.
  exists M. exists M0.
  split; auto.
  unfold MemMod.disjUnion. split; auto.
Qed.
Hint Resolve mstar_mwand_equ_left.

Lemma mstar_mwand_equ_right:
  forall m m' p S, pm_mwand m (pm_mwand m' p) S -> pm_mwand (mstar m m') p S.
Proof.
  intros.
  unfold pm_mwand.
  unfold pm_mwand in H.
  intros.
  induction S. induction a. induction a.
  rename a into MS. rename b1 into R. rename b0 into ie. rename b into is.
  UnfoldStateAll.
  unfold mstar in H1.
  InvertAll.
  rename x into M1. rename x0 into M2.
  assert (MemMod.disjoint M1 MS).
  unfold MemMod.disjUnion in H2. inversion_clear H2.
  rewrite H5 in H0. unfold MemMod.disjUnion in H0.
  inversion_clear H0.
  eauto with PMap.
  apply H with M1 (MemMod.merge M1 MS) M2; auto.
  unfold MemMod.disjUnion. split; auto.
  assert (MemMod.disjUnion M2 M1 M).
  auto with PMap.
  unfold MemMod.disjUnion in H5. inversion_clear H5.
  rewrite H7 in H0. auto with PMap.
Qed.
Hint Resolve mstar_mwand_equ_right.

Lemma mstar_commut:
  forall m m' M, mstar m m' M -> mstar m' m M.
Proof.
  unfold mstar.
  intros.
  InvertAll.
  exists x0. exists x.
  split; auto with PMap.
Qed.
Hint Resolve mstar_commut.

Lemma mstar_sym : forall M A B,
  (A *^* B) M -> (B *^* A) M. 
Proof.  
  intros M A B [ha [hb [J [Ha Hb]]]].
  exists hb; exists ha.
  split.
    auto with PMap.
  auto.
Qed.

Lemma mstar_assoc_left_right:
  forall M A B C,
    (A *^* B *^* C) M
    -> (A *^* (B *^* C)) M.
Proof.
  intros.
  destruct H as [hab [hc [Jabc [HAB HC]]]].
  destruct HAB as [ha [hb [Jab [HA HB]]]].
  exists ha; exists (MemMod.merge hb hc).
  split.
    apply MemMod.disjUnion_assoc_left_right.
    destruct Jab; trivial.
    elim Jab.
    intros.
    rewrite <- H0; trivial.
  split_l.
  exists hb; exists hc; split; auto.
  split_r.
  destruct Jab.
  destruct Jabc.
  subst hab.
  apply MemMod.disj_merge_disj_right with ha; eauto.
Qed.
Implicit Arguments mstar_assoc_left_right [M A B C].

Lemma mstar_assoc_right_left:
  forall M A B C,
    (A *^* (B *^* C)) M
    -> (A *^* B *^* C) M.
Proof.
  intros.
  apply mstar_sym.
  apply mstar_assoc_left_right.
  exact (mstar_sym (mstar_assoc_left_right (mstar_sym H))).
Qed.
Implicit Arguments mstar_assoc_left_right [M A B C].

Lemma leadsToInterK:
  forall sp h, leadsToBlk sp 6 h
    -> exists pc, leadsToW sp pc h.
Proof.
intros.
unfold leadsToBlk in H.
unfold ptoxBlk in H.
unfold imprecM in H.
unfold mstar in H.
InvertAll.
unfold ptox in H.
InvertAll.
exists x13.

unfold leadsToW.
unfold ptoW.
unfold imprecM.
unfold mstar.

unfold MemMod.disjUnion in H1.
unfold MemMod.disjUnion in H3.
InvertAll.

exists (MemMod.merge x1 x3).
exists (MemMod.merge x0 x4).

split.
apply MemMod.disjUnion_commut.
apply MemMod.disjUnion_assoc_right_left.
apply MemMod.disjoint_commut.
apply MemMod.disj_merge_assoc_right_left.
auto.
rewrite <- H15; auto.

assert (MemMod.merge x3 x1 = MemMod.merge x1 x3).
apply MemMod.disj_merge_commut; auto.
apply MemMod.disj_merge_disj_left with x4.
rewrite <- H15.
apply MemMod.disjoint_commut; auto.
rewrite <- H1.

assert (MemMod.merge (MemMod.merge x3 x4) x1 = MemMod.merge x4 (MemMod.merge x3 x1)).
apply MemMod.disj_merge_ABC_BAC; auto.
rewrite <- H17.
rewrite <- H15.
assert (MemMod.merge x1 x2 = MemMod.merge x2 x1).
apply MemMod.disj_merge_commut; auto.
rewrite <- H18.
rewrite <- H16.
apply MemMod.disjUnion_commut; auto.

split.
 exists x1.
 exists x3.
 split.
  unfold MemMod.disjUnion.
  split.
  apply MemMod.disjoint_commut.
  apply MemMod.disj_merge_disj_left with x4.
  rewrite <- H15.
  apply MemMod.disjoint_commut; auto. auto.

  split; auto.
unfold mTrue; auto.
Qed.

Lemma ptoW_indom:
  forall l m h,
  ptoW l m h
  -> MemMod.in_dom l h /\ MemMod.in_dom (l+1) h.
Proof.
unfold ptoW.
intros l m h pW. 
destruct pW as [M1 [M2 [djh  [mM1 mM2]]]].
unfold ptox in mM2.
unfold pto in *.
unfold MemMod.sgltonM in *.
assert (MemMod.in_dom l M1 /\ MemMod.in_dom (l+1) M2).
unfold MemMod.in_dom in |- *.
unfold MemMod.maps_to in *.
InvertAll.
split.
exists m; auto.
exists x; auto.
unfold MemMod.disjUnion in djh.
InvertAll.
rewrite H4.
split.
apply MemMod.merge_presv_indom_left; auto.
apply MemMod.merge_presv_indom_right; auto.
Qed.


Lemma ptoxBlk_notindom: forall l s addr h,
  ptoxBlk l s h
  -> addr < l \/ addr >= l + s
  -> h addr = None.
Proof.
intros l s.
gen_clear l.
induction s; intros.
simpl in H.
unfold Aemp in H.
subst;trivial.
simpl in H.
destruct H as [h1 [h2 [HJh H]]].
destruct H.
assert (Hx : addr < l + 1 \/ addr >= l+1+s).
  omega.
assert (Hx0 := IHs (l+1) addr h2 H1 Hx).
assert (Hx1 : h1 addr = None).
unfold ptox in H.
inversion H.
unfold pto in H2.
unfold MemMod.sgltonM in H2.
destruct H2.
unfold MemMod.not_in_dom in H3.
apply H3.
omega.
unfold MemMod.disjUnion in HJh.
destruct HJh.
unfold MemMod.merge in H3.
subst h.
rewrite Hx1.
trivial.
Qed.

Lemma ptoxBlk_split : forall s1 s2 M l,
  ptoxBlk l (s1 + s2) M
  -> mstar (ptoxBlk l s1) (ptoxBlk (l+s1) s2) M.
Proof.
  induction s1; simpl; intros.
  unfold Aemp, MemMod.emp.
  exists MemMod.emp; exists M.
  split.
    auto with PMap.
    unfold MemMod.emp.
    split_l.
  replace (l+0) with l; auto with arith.
  destruct  H as [M1 [M2 [MJM [HM1 HM2]]]].
  replace (l + S s1) with (S l + s1); [idtac | omega].
  assert (Hx := IHs1 s2 M2 (l+1) HM2).
  apply mstar_assoc_right_left.
  unfold mstar.
  exists M1; exists M2; split_l.
  split_l.
  arith_rep (S l + s1) with (l+1+s1).
Qed.

Lemma ptoxBlk_merge : forall s1 s2 M l,
  mstar (ptoxBlk l s1) (ptoxBlk (l+s1) s2) M
  -> ptoxBlk l (s1 + s2) M.
Proof.
induction s1;simpl;intros.
unfold mstar, Aemp in H.
InvertAll.
subst.
assert (x0 = M).
auto with PMap.
subst.
arith_rep l with (l+0);assumption.
assert ( (ptox l *^* (ptoxBlk (l + 1) s1 *^* ptoxBlk (l + S s1) s2)) M).
apply mstar_assoc_left_right; assumption.
unfold mstar at 1 in H0.
InvertAll.
exists x;exists x0;intuition.
apply IHs1.
arith_rep (l+1+s1) with (l+S s1);assumption.
Qed.

Lemma stkSpacePrec_split : forall sp s1 s2 M,
  stkSpacePrec sp (s1 + s2) M
  -> mstar (stkSpacePrec sp s2) (stkSpacePrec (sp-s2) s1) M.
Proof.
unfold stkSpacePrec.
intros.
destruct H.
assert ((ptoxBlk (sp -(s1+s2)) s1  *^* ptoxBlk (sp-(s1+s2)+s1) s2 ) M).
apply ptoxBlk_split;assumption.
unfold mstar in H1;InvertAll.
exists x0;exists x;intuition.
arith_rep (sp-s2) with (sp-(s1+s2)+s1).
arith_rep (sp-s2-s1) with (sp-(s1+s2)).
Qed.
Implicit Arguments stkSpacePrec_split [sp s1 s2 M].


Lemma stkSpacePrec_merge : forall sp s1 s2 M1 M2,
  stkSpacePrec sp s2 M1
  -> stkSpacePrec (sp-s2) s1 M2
  -> stkSpacePrec sp (s1 + s2) (MemMod.merge M1 M2).
Proof.
unfold stkSpacePrec.
intros.
InvertAll.
intuition.
apply ptoxBlk_merge.
assert (MemMod.disjoint M2 M1).
unfold MemMod.disjoint.
unfold MemMod.not_in_dom.
intros.
destruct (le_lt_dec (sp-s2) x).
left;apply ptoxBlk_notindom with (sp-s2-s1) s1.
assumption.
right;omega.
right;apply ptoxBlk_notindom with (sp-s2) s2.
assumption.
left;assumption.
exists M2;exists M1;intuition.
arith_rep (sp-(s1+s2)) with (sp-s2-s1).
arith_rep (sp-(s1+s2)+s1) with (sp-s2).
Qed.
Implicit Arguments stkSpacePrec_merge [sp s1 s2 M1 M2].

Definition uniqueM (m: MPred) : Prop :=
forall M M', m M -> m M' -> M = M'.

