(***************************************************************)
(* Basci Coq lib                                               *)
(***************************************************************)
Require Import Bool.

(** Section 1: optionT and pairT **)
Axiom extensionality:
  forall (A B: Set) (f g : A -> B),
  (forall x, f x = g x) -> f = g.

Inductive OptionT (T : Type) : Type :=
    NoneT1 : OptionT T
  | SomeT1 : T -> OptionT T.

Inductive optionT (A : Type) : Type :=
  | SomeT : A -> optionT A
  | NoneT : optionT A.

Implicit Arguments SomeT [A].
Implicit Arguments NoneT [A].

Notation opt_predT :=
  (fun opt:optionT _ => 
    fun f =>
      match opt with
	| SomeT t => f t
	| NoneT => False
      end).

Notation optpT :=
  (fun f => fun opt:optionT _ => 
    match opt with
      | SomeT t => f t
      | NoneT => False
    end).

Notation optgT :=
  (fun f => fun opt:optionT _ =>
    fun t' =>
      match opt with
        | SomeT t => f t t'
        | NoneT => False
      end).

Notation opt_predT2 :=
  (fun opt:optionT _ => 
    fun opt':optionT _ => 
      fun f =>
	match opt, opt' with
	  | SomeT t, SomeT t' => f t t'
	  | _, _ => False
	end).

Inductive prodT (A : Set) (B : Type) : Type :=
    pairT : A -> B -> prodT A B.
Implicit Arguments pairT [A B].

Notation "( x , y , .. , z )" :=
  (pairT .. (pairT x y) .. z) : t_type_scope.

(** Section 2: Tactics **)
(***************************************************************)

Ltac hypreplace H y :=
  let H' := fresh in
    (assert (H' := y); try (clear H; rename H' into H)).

Ltac hypreplace2 H y Hn:=
  let H' := fresh in
    (assert (H' := y); try (clear H; rename H' into Hn)).

Tactic Notation "substH" hyp (H) :=
  match goal with
    | |- ?a -> ?b => clear H; intro H
    | |- _ => fail 1 "goal must be 'A -> B'."
  end.

Tactic Notation "substH" hyp (H) "with" constr (t) := 
  hypreplace H t.

Tactic Notation "substH" hyp (H) "with" constr (t) "into" ident (Hn) := 
  hypreplace2 H t Hn.

Tactic Notation "bool_destruct" hyp (H) "as" simple_intropattern (pat) :=
  let H0 := fresh "H" in
    (rename H into H0;
    match type of (H0) with
      | (andb ?a ?b = true) => 
        destruct (andb_prop a b H0) as pat; clear H0
      | (orb ?a ?b = true) =>
        destruct (orb_prop a b H0) as pat; clear H0
      | _ => fail "not destructable" 
    end).


Tactic Notation "destructH" hyp (H) "with" constr (t) :=
  destruct t; clear H.

Tactic Notation "destructH" hyp (H) "with" constr (t)
  "as" simple_intropattern (p) :=
  destruct t as p; clear H.
  
Tactic Notation "discri" :=
  match goal with
    | H : ?a <> ?a |- _ => 
      elimtype False; apply H; reflexivity
    | |- _ -> _ => intros; discriminate
    | H : False |- _ => destruct H
    | _ => discriminate
  end.

Tactic Notation "gen_clear" hyp (H) :=
  generalize H; clear H.

Tactic Notation "gen_clear" hyp (H1) hyp (H2) :=
  generalize H1 H2; clear H1 H2.

Tactic Notation "gen_clear" 
  hyp (H1) hyp (H2) hyp (H3) :=
  generalize H1 H2 H3; clear H1 H2 H3.

Tactic Notation "gen_clear" 
  hyp (H1) hyp (H2) hyp (H3) hyp (H4) :=
  generalize H1 H2 H3 H4; 
    clear H1 H2 H3 H4.

Tactic Notation "gen_clear" 
  hyp (H1) hyp (H2) hyp (H3) 
  hyp (H4) hyp (H5):=
  generalize H1 H2 H3 H4 H5; 
    clear H1 H2 H3 H4 H5.

Tactic Notation "gen_clear" 
  hyp (H1) hyp (H2) hyp (H3) 
  hyp (H4) hyp (H5) hyp (H6):=
  generalize H1 H2 H3 H4 H5 H6; 
    clear H1 H2 H3 H4 H5 H6.

Tactic Notation "split_l" :=
  split; [trivial | idtac].

Tactic Notation "split_r" :=
  split; [idtac | trivial ].

Tactic Notation "split_lr" :=
  split; [trivial | trivial ].

Tactic Notation "split_l" "with" constr (t) :=
  split; [apply t | idtac].

Tactic Notation "split_r" "with" constr (t) :=
  split; [idtac | apply t ].

Tactic Notation "split_l" "by" tactic (tac) :=
  split; [tac | idtac ].

Tactic Notation "split_r" "by" tactic (tac) :=
  split; [idtac | tac ].

Tactic Notation "split_l_clear" "with" hyp (H) :=
  split; [apply H | clear H].

Tactic Notation "split_r_clear" "with" hyp (H) :=
  split; [clear H | apply H ].

Lemma and_sym : forall (A B : Prop), A /\ B -> B /\ A.
Proof. intros  A B [HA HB]; split; trivial. Qed.
Implicit Arguments and_sym [A B].

Ltac rsplit := apply and_sym; split.

Lemma and_sym_rr : forall A B C : Prop, A /\ B /\ C -> B /\ C /\ A.
Proof.
  tauto.
Qed.

Ltac rrsplit := apply and_sym_rr; split.

Tactic Notation "inj_hyp" hyp (H) :=
  injection H; clear H; intro H.

Tactic Notation "rew_clear" hyp (H) :=
  rewrite H; clear H.

Tactic Notation "injection" hyp (H) :=
  injection H.

Tactic Notation "injection" hyp (H) "as" 
  simple_intropattern (pat) :=
  injection H; intros pat.

Tactic Notation "injsubst" ident (id) "in" hyp (H) :=
  injection H; intro; subst id; clear H.

Ltac InvertAll :=
  repeat
    match goal with
      | H: _ /\ _ |- _ => inversion_clear H
      | H: ex _   |- _ => inversion_clear H
    end.

Ltac arith_replace t1 t2 := 
  (replace t1 with t2; fail "error") ||
    (replace t1 with t2; [trivial | try omega; fail "error" ]).

Ltac arith_replaceH H t1 t2 := 
  (replace t1 with t2 in H; fail "error") ||
    (replace t1 with t2 in H; [trivial | try omega; fail "error" ]).

Tactic Notation "arith_rep" constr(t1) "with" constr (t2) :=
  arith_replace t1 t2.

Tactic Notation "arith_rep" constr(t1) "with" constr (t2) "in" hyp (H):=
  arith_replaceH H t1 t2.


Ltac clearall := 
  match goal with 
    | H : _ |- _ =>
      (clear H || (generalize H; clear H)); clearall
    | _ => intros
  end.

Ltac clearall_arith := 
  match goal with 
    | H : ?a > ?b |- _ => (generalize H; clear H); clearall_arith
    | H : ?a >= ?b |- _ => (generalize H; clear H); clearall_arith
    | H : ?a < ?b |- _ => (generalize H; clear H); clearall_arith
    | H : ?a <= ?b |- _ => (generalize H; clear H); clearall_arith
    | H : ?a = ?b |- _ => 
      match type of a with
        | nat => (generalize H; clear H); clearall_arith
        | _ => (clear H || (generalize H; clear H)); clearall_arith
      end
    | H : _ |- _ =>
      (clear H || (generalize H; clear H)); clearall_arith
    | _ => intros
  end.

(* *********************************************************************** *)
(** * The [skip] tactic from Stephanie Weirich and Brian Andemir of UPenn  *)

(** The [skip] tactic will prove any goal, since we assert below that
    [False] is provable.  This is a useful technique for skipping over
    troublesome parts of proofs.  In a complete development, one would
    comment out the [skip] axiom and tactic.

    In the exercises below, [skip] lets us give partial proofs by
    using it as a placeholder for locations in the proof script where
    the reader should fill in the proof. *)

