(***************************************************************)
(* Coq code for PCAP                                           *)
(*                                                             *)
(* List of Type                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Require Import Le.
Require Import Coqlib.

Section ListT.

  Variable A : Type.
  
  Set Implicit Arguments.

  Inductive listT : Type :=
    | nilT : listT
    | consT : A -> listT -> listT.
  
  Infix "::" := consT 
    (at level 60, right associativity) : listt_scope.
  
  Open Local Scope listt_scope.
  
  (*************************)
  (** Discrimination       *)
  (*************************)
  
  Lemma nilT_consT : forall (a : A) (m : listT), nilT <> a :: m.
  Proof. 
    intros; discriminate.
  Qed.
  
  (*************************)
  (** Concatenation        *)
  (*************************)
  
  Fixpoint appT (l m : listT) {struct l} : listT:=
    match l with
      | nilT => m
      | a :: l1 => a :: appT l1 m
    end.
  
  Infix "++" := appT (right associativity, at level 60) : listt_scope.
  
  Lemma appT_nilT_end : forall l : listT, l = l ++ nilT.
  Proof. 
    induction l; simpl in |- *; auto.
    rewrite <- IHl; auto.
  Qed.
  Hint Resolve appT_nilT_end.
  
  Ltac now_show c := change c in |- *.
  
  Lemma appT_ass : forall l m n : listT, (l ++ m) ++ n = l ++ m ++ n.
  Proof. 
    intros. induction l; simpl in |- *; auto.
    now_show (a :: (l ++ m) ++ n = a :: l ++ m ++ n).
    rewrite <- IHl; auto.
  Qed.
  Hint Resolve appT_ass.
  
  Lemma ass_appT : forall l m n:listT, l ++ m ++ n = (l ++ m) ++ n.
  Proof. 
    auto.
  Qed.
  Hint Resolve ass_appT.
  
  Lemma appT_comm_consT : forall (x y:listT) (a:A), a :: x ++ y = (a :: x) ++ y.
  Proof.
    auto.
  Qed.
  
  Lemma appT_eq_nilT : forall x y:listT, x ++ y = nilT -> x = nilT /\ y = nilT.
  Proof.
    destruct x as [| a l]; [ destruct y as [| a l] | destruct y as [| a0 l0] ];
      simpl in |- *; auto.
    intros H; discriminate H.
    intros; discriminate H.
  Qed.
  
  Lemma appT_consT_not_nilT : forall (x y:listT) (a:A), nilT <> x ++ a :: y.
  Proof.
    unfold not in |- *.
    destruct x as [| a l]; simpl in |- *; intros.
    discriminate H.
    discriminate H.
  Qed.
  
Lemma appT_eq_unit :
 forall (x y:listT) (a:A),
   x ++ y = a :: nilT -> x = nilT /\ y = a :: nilT \/ x = a :: nilT /\ y = nilT.

Proof.
  destruct x as [| a l]; [ destruct y as [| a l] | destruct y as [| a0 l0] ];
   simpl in |- *.
  intros a H; discriminate H.
  left; split; auto.
  right; split; auto.
  generalize H.
  generalize (appT_nilT_end l); intros E.
  rewrite <- E; auto.
  intros.
  injection H.
  intro.
  cut (nilT = l ++ a0 :: l0); auto.
  intro.
  generalize (appT_consT_not_nilT _ _ _ H1); intro.
  elim H2.
Qed.

Lemma appT_inj_tail :
 forall (x y:listT) (a b:A), x ++ a :: nilT = y ++ b :: nilT -> x = y /\ a = b.
Proof.
  induction x as [| x l IHl];
   [ destruct y as [| a l] | destruct y as [| a l0] ]; 
   simpl in |- *; auto.
  intros a b H.
  injection H.
  auto.
  intros a0 b H.
  injection H; intros.
  generalize (appT_consT_not_nilT _ _ _ H0); destruct 1.
  intros a b H.
  injection H; intros.
  cut (nilT = l ++ a :: nilT); auto.
  intro.
  generalize (appT_consT_not_nilT _ _ _ H2); destruct 1.
  intros a0 b H.
  injection H; intros.
  destruct (IHl l0 a0 b H0). 
  split; auto.
  rewrite <- H1; rewrite <- H2; reflexivity.
Qed.

(*************************)
(** Head and tail        *)
(*************************)

Definition head (l : listT) :=
  match l return optionT A with
  | nilT => NoneT
  | x :: _ => SomeT x
  end.

Definition tail (l:listT) : listT:=
  match l with
  | nilT => nilT
  | a :: m => m
  end.

(****************************************)
(** Length of lists                     *)
(****************************************)

Fixpoint lengthT (l:listT) : nat :=
  match l with
  | nilT => 0
  | _ :: m => S (lengthT m)
  end.

(******************************)
(** LengthT order of lists     *)
(******************************)

Section lengthT_order.
Definition lel (l m:listT) := lengthT l <= lengthT m.

Variables a b : A.
Variables l m n : listT.

Lemma lel_refl : lel l l.
Proof. 
	unfold lel in |- *; auto with arith.
Qed.

Lemma lel_trans : lel l m -> lel m n -> lel l n.
Proof. 
	unfold lel in |- *; intros.
        now_show (lengthT l <= lengthT n).
        apply le_trans with (lengthT m); auto with arith.
Qed.

Lemma lel_consT_consT : lel l m -> lel (a :: l) (b :: m).
Proof. 
	unfold lel in |- *; simpl in |- *; auto with arith.
Qed.

Lemma lel_consT : lel l m -> lel l (b :: m).
Proof. 
	unfold lel in |- *; simpl in |- *; auto with arith.
Qed.

Lemma lel_tail : lel (a :: l) (b :: m) -> lel l m.
Proof. 
	unfold lel in |- *; simpl in |- *; auto with arith.
Qed.

Lemma lel_nilT : forall l':listT, lel l' nilT -> nilT = l'.
Proof. 
	intro l'; elim l'; auto with arith.
	intros a' y H H0.
        now_show (nilT = a' :: y).
	absurd (S (lengthT y) <= 0); auto with arith.
Qed.
End lengthT_order.

Hint Resolve lel_refl lel_consT_consT lel_consT lel_nilT lel_nilT nilT_consT.

(*********************************)
(** The [In] predicate           *)
(*********************************)

Fixpoint In (a:A) (l:listT) {struct l} : Prop :=
  match l with
  | nilT => False
  | b :: m => b = a \/ In a m
  end.

Lemma in_eq : forall (a:A) (l:listT), In a (a :: l).
Proof. 
  simpl in |- *; auto.
Qed.
Hint Resolve in_eq.

Lemma in_consT : forall (a b:A) (l:listT), In b l -> In b (a :: l).
Proof. 
	simpl in |- *; auto.
Qed.
Hint Resolve in_consT.

Lemma in_nilT : forall a:A, ~ In a nilT.
Proof.
  unfold not in |- *; intros a H; inversion_clear H.
Qed.


Lemma in_inv : forall (a b:A) (l:listT), In b (a :: l) -> a = b \/ In b l.
Proof.
 intros a b l H; inversion_clear H; auto.
Qed.

Lemma In_dec :
 (forall x y:A, {x = y} + {x <> y}) ->
 forall (a:A) (l:listT), {In a l} + {~ In a l}.

Proof.
  intro H.
  induction l as [| a0 l IHl].
  right; apply in_nilT.
  destruct (H a0 a); simpl in |- *; auto.
  destruct IHl; simpl in |- *; auto. 
  right; unfold not in |- *; intros [Hc1| Hc2]; auto.
Qed.

Lemma in_appT_or : forall (l m:listT) (a:A), In a (l ++ m) -> In a l \/ In a m.
Proof. 
	intros l m a.
	elim l; simpl in |- *; auto.
	intros a0 y H H0.
	now_show ((a0 = a \/ In a y) \/ In a m).
	elim H0; auto.
	intro H1.
        now_show ((a0 = a \/ In a y) \/ In a m).
	elim (H H1); auto.
Qed.
Hint Immediate in_appT_or.

Lemma in_or_appT : forall (l m:listT) (a:A), In a l \/ In a m -> In a (l ++ m).
Proof. 
	intros l m a.
	elim l; simpl in |- *; intro H.
        now_show (In a m).
	elim H; auto; intro H0.
	now_show (In a m).
	elim H0. (* subProof completed *)
	intros y H0 H1.
	now_show (H = a \/ In a (y ++ m)).
	elim H1; auto 4.
	intro H2.
	now_show (H = a \/ In a (y ++ m)).
	elim H2; auto.
Qed.
Hint Resolve in_or_appT.

(***************************)
(** Set inclusion on listT *)
(***************************)

Definition incl (l m:listT) := forall a:A, In a l -> In a m.
Hint Unfold incl.

Lemma incl_refl : forall l:listT, incl l l.
Proof. 
	auto.
Qed.
Hint Resolve incl_refl.

Lemma incl_tl : forall (a:A) (l m:listT), incl l m -> incl l (a :: m).
Proof. 
	auto.
Qed.
Hint Immediate incl_tl.

Lemma incl_tran : forall l m n:listT, incl l m -> incl m n -> incl l n.
Proof. 
	auto.
Qed.

Lemma incl_appTl : forall l m n:listT, incl l n -> incl l (n ++ m).
Proof. 
	auto.
Qed.
Hint Immediate incl_appTl.

Lemma incl_appTr : forall l m n:listT, incl l n -> incl l (m ++ n).
Proof. 
	auto.
Qed.
Hint Immediate incl_appTr.

Lemma incl_consT :
 forall (a:A) (l m:listT), In a m -> incl l m -> incl (a :: l) m.
Proof. 
	unfold incl in |- *; simpl in |- *; intros a l m H H0 a0 H1.
        now_show (In a0 m).
	elim H1.
	now_show (a = a0 -> In a0 m).
	elim H1; auto; intro H2.
	now_show (a = a0 -> In a0 m).
	elim H2; auto. (* solves subgoal *)
	now_show (In a0 l -> In a0 m).
	auto.
Qed.
Hint Resolve incl_consT.

Lemma incl_appT : forall l m n:listT, incl l n -> incl m n -> incl (l ++ m) n.
Proof. 
	unfold incl in |- *; simpl in |- *; intros l m n H H0 a H1.
        now_show (In a n).
	elim (in_appT_or _ _ _ H1); auto.
Qed.
Hint Resolve incl_appT.

(**************************)
(** Nth element of a listT*)
(**************************)

Fixpoint nth (n:nat) (l:listT) (default:A) {struct l} : A :=
  match n, l with
  | O, x :: l' => x
  | O, other => default
  | S m, nilT => default
  | S m, x :: t => nth m t default
  end.

Fixpoint nth_ok (n:nat) (l:listT) (default:A) {struct l} : bool :=
  match n, l with
  | O, x :: l' => true
  | O, other => false
  | S m, nilT => false
  | S m, x :: t => nth_ok m t default
  end.

Lemma nth_in_or_default :
 forall (n:nat) (l:listT) (d:A), {In (nth n l d) l} + {nth n l d = d}.
(* Realizer nth_ok. Program_all. *)
Proof. 
  intros n l d; generalize n; induction l; intro n0.
  right; case n0; trivial.
  case n0; simpl in |- *.
  auto.
  intro n1; elim (IHl n1); auto.     
Qed.

Lemma nth_S_consT :
 forall (n:nat) (l:listT) (d a:A),
   In (nth n l d) l -> In (nth (S n) (a :: l) d) (a :: l).
Proof. 
  simpl in |- *; auto.
Qed.


(********************************)
(** Decidable equality on lists *)
(********************************)


Lemma list_eq_dec :
 (forall x y:A, {x = y} + {x <> y}) -> forall x y:listT, {x = y} + {x <> y}.
Proof.
  intro H.
  induction x as [| a l IHl]; destruct y as [| a0 l0]; auto.
  destruct (H a a0) as [e| e].
  destruct (IHl l0) as [e'| e'].
  left; rewrite e; rewrite e'; trivial.
  right; red in |- *; intro.
  apply e'; injection H0; trivial.
  right; red in |- *; intro.
  apply e; injection H0; trivial.
Qed.

(*************************)
(**  Reverse             *)
(*************************)

Fixpoint rev (l:listT) : listT:=
  match l with
  | nilT => nilT
  | x :: l' => rev l' ++ x :: nilT
  end.

Lemma distr_rev : forall x y:listT, rev (x ++ y) = rev y ++ rev x.
Proof.
 induction x as [| a l IHl].
 destruct y as [| a l].
 simpl in |- *.
 auto.

 simpl in |- *.
 apply appT_nilT_end; auto.

 intro y.
 simpl in |- *.
 rewrite (IHl y).
 apply (appT_ass (rev y) (rev l) (a :: nilT)).
Qed.

Remark rev_unit : forall (l:listT) (a:A), rev (l ++ a :: nilT) = a :: rev l.
Proof.
 intros.
 apply (distr_rev l (a :: nilT)); simpl in |- *; auto.
Qed.

Lemma rev_involutive : forall l:listT, rev (rev l) = l.
Proof.
 induction l as [| a l IHl].
 simpl in |- *; auto.

 simpl in |- *.
 rewrite (rev_unit (rev l) a).
 rewrite IHl; auto.
Qed.

(*********************************************)
(**  Reverse Induction Principle on Lists    *)
(*********************************************)

Section Reverse_Induction.

Unset Implicit Arguments.

Remark rev_list_ind :
 forall P:listT-> Prop,
   P nilT ->
   (forall (a:A) (l:listT), P (rev l) -> P (rev (a :: l))) ->
   forall l:listT, P (rev l).
Proof.
 induction l; auto.
Qed.
Set Implicit Arguments.

Lemma rev_ind :
 forall P:listT-> Prop,
   P nilT ->
   (forall (x:A) (l:listT), P l -> P (l ++ x :: nilT)) -> forall l:listT, P l.
Proof.
 intros.
 generalize (rev_involutive l).
 intros E; rewrite <- E.
 apply (rev_list_ind P).
 auto.

 simpl in |- *.
 intros.
 apply (H0 a (rev l0)).
 auto.
Qed.

End Reverse_Induction.

End ListT.

Implicit Arguments nilT [A].

Hint Resolve nilT_consT appT_nilT_end ass_appT appT_ass: datatypes v62.
Hint Resolve appT_comm_consT appT_consT_not_nilT: datatypes v62.
Hint Immediate appT_eq_nilT: datatypes v62.
Hint Resolve appT_eq_unit appT_inj_tail: datatypes v62. 
Hint Resolve lel_refl lel_consT_consT lel_consT lel_nilT lel_nilT nilT_consT:
  datatypes v62.
Hint Resolve in_eq in_consT in_inv in_nilT in_appT_or in_or_appT: datatypes v62.
Hint Resolve incl_refl incl_tl incl_tran incl_appTl incl_appTr incl_consT
  incl_appT: datatypes v62.

(** Exporting listTnotations *)

Infix "::" := consT (at level 60, right associativity) : listt_scope.

Infix "++" := appT (right associativity, at level 60) : listt_scope.

Open Scope listt_scope.

(** Declare Scope list_scope with key listT*)
Delimit Scope listt_scope with listT.

Bind Scope listt_scope with listT.
