Require Import Coqlib.
Require Import Map.
Require Import Arith.

Set Implicit Arguments.
Unset Strict Implicit.

Notation   Label := nat   (only parsing).
Notation   int   := nat   (only parsing).

Module MemSpec <: MapSpec.
  Definition domain := Label.
  Definition image := int.
  Definition beq_Domain := eq_nat_dec.
End MemSpec.

Module MemMod := MapFun(MemSpec).

Definition Mem := MemMod.Map.


(** cell l in Mem *)
Definition l_inH (l : Label) (h : Mem) : Prop :=
  MemMod.in_dom l h.
  
(** Store w into Mem l *)
Definition H_Set (h : Mem) (l : Label) (w : int) : Mem :=
  MemMod.put h l w.
 
(** load w from Mem l *)
Definition H_Get (h : Mem) (l : Label) : int :=
  match h l with
    | Some b => b
    | None => 0
  end.

Inductive reg : Set := AX : reg
                     | BX : reg
                     | CX : reg
                     | DX : reg
                     | BP : reg
                     | SI : reg
                     | DI : reg
                     | SP : reg.


Definition beq_reg (r r': reg) :=
  match r, r' with
  | AX, AX => true
  | BX, BX => true
  | CX, CX => true
  | DX, DX => true
  | BP, BP => true
  | DI, DI => true
  | SI, SI => true
  | SP, SP => true
  | _,  _  => false
  end.


Definition RFile := reg -> nat.

Definition RF_update (R : RFile) (r : reg) (n : nat) : RFile :=
  fun (r' : reg) =>
    if beq_reg r' r then n else R r'.

Inductive flag : Set :=
    _ZF : flag
  | _IF : flag.

Definition beq_flag(f f' : flag) :=
  match f, f' with
  | _ZF, _ZF => true
  | _IF, _IF => true
  | _, _ => false
  end.

Definition FlagR := flag -> bool.

Definition Set_Flag (FL : FlagR) (fl : flag) : FlagR :=
  fun (fl' : flag) =>
    if beq_flag fl' fl then true else FL fl'.

Definition Clear_Flag (FL : FlagR) (fl : flag) : FlagR :=
  fun (fl' : flag) =>
    if beq_flag fl' fl then false else FL fl'.

Parameter FR2nat : FlagR -> nat.
Parameter nat2FR : nat -> FlagR.
Parameter FR_nat_iso_left : forall fr : FlagR, nat2FR (FR2nat fr) = fr.


(* ISR is a special register in the PIC          *)
(* here we do not model the full PIC             *)
(* instead, we treat it as a special reg in CPU  *)
(* Since we only support one interrupt,          *)
(* it is sufficient to model it as boolean value *)
Notation ISR := bool    (only parsing).



(*******************************************************************)
(** pc, program counter                                            *)
Definition PC := int.

(** entry point for the timer interrupt handler *)
Parameter  timer : PC.
Definition h_entry := timer.


Definition State := (Mem * RFile * FlagR * ISR)%type.


(** Machine parameter manipulating                                *)
(** A_B, meams, find "Part B" in "A"                              *)
Definition S_Mem (S : State) : Mem :=
  match S with
    (M, _, _, _) => M
  end.

Definition S_RF (S : State) : RFile :=
  match S with
    (_, R, _, _) => R
  end.

Definition S_Flag (S : State) : FlagR :=
  match S with
    (_, _, fl, _) => fl
  end.

Definition S_ISR (S : State) : ISR :=
  match S with
    (_, isr) => isr
  end.


(** Update Machine State *******************************************)
(** A_B_Upd, meams, update "Part B" in "A"                         *)
Definition S_Mem_Upd (S : State) (M : Mem) : State :=
  (M, S_RF S, S_Flag S, S_ISR S).

Definition S_RF_Upd (S : State) (R : RFile) : State :=
  (S_Mem S, R, S_Flag S, S_ISR S).

Definition S_Flag_Upd (S : State) (fl : FlagR) : State :=
  (S_Mem S, S_RF S, fl, S_ISR S).

Definition S_ISR_Upd (S : State) (isr : ISR) : State :=
  (S_Mem S, S_RF S, S_Flag S, isr).

Tactic Notation "simplS" "in" hyp(H) := 
  unfold S_Mem_Upd, S_RF_Upd, S_Flag_Upd, S_ISR_Upd, 
    S_Mem, S_RF, S_Flag, S_ISR in H.

Tactic Notation "simplS" := 
  unfold S_Mem_Upd, S_RF_Upd, S_Flag_Upd, S_ISR_Upd, 
    S_Mem, S_RF, S_Flag, S_ISR.

Tactic Notation "simplS" "in" "*" := 
  unfold S_Mem_Upd, S_RF_Upd, S_Flag_Upd, S_ISR_Upd, 
    S_Mem, S_RF, S_Flag, S_ISR in * .

(** defination of labels *)
Notation  lab := Label (only parsing).
Notation flab := Label (only parsing).

(* address mode *)
Inductive Addr : Set :=
 | A_imm : int -> Addr
 | A_reg : int -> reg -> Addr.

(* oprand *)
Inductive Opr : Set :=
 | Opr_imm : int -> Opr
 | Opr_reg : reg -> Opr.
  
(** Common Instructions: l *)
Inductive instr : Set :=
 | _cli     : instr
 | _sti     : instr
  (** Function Call instruction **)
 | _call    : lab -> instr
  (** Branch instruction **)
 | _jne    : lab -> instr
  (** mov instructions **)
 | _movwp   : Opr -> reg -> instr
 | _movwld  : Addr -> reg -> instr
 | _movwst  : Opr -> Addr -> instr
  (** stack operations **)
 | _pushwp  : Opr -> instr
 | _pushwm  : Addr -> instr
 | _pushfw  : instr
 | _popw    : reg -> instr
 | _popfw   : instr
  (** Simple instructions **)
 | _cmpw    : Opr -> reg -> instr
 | _addw    : Opr -> reg -> instr
 | _subw    : Opr -> reg -> instr
  (** a special instruction for eoi **)
 | _outb_eoi: instr.          
 (* outb_eoi is a hardwired instruction corresponding to
    outb al dx, where both al and dx contain 0x20 *)


(*********************************
  We would like to support the following
  mov instructions:
 | _movir   : int -> reg -> instr            (* movw n r     *)
 | _movrr   : reg -> reg -> instr            (* movw r r'    *)
 | _movldi  : int -> reg -> instr            (* movw (n) r   *)
 | _movldr  : int -> reg -> reg -> instr     (* movw n(r) r' *)
   (* store registers into memory *)
 | _movstri : reg -> int -> instr            (* movw r (n)   *)
 | _movstrr : reg -> int -> reg -> instr     (* movw r n(r') *)
   (* store immediate num into memory *)
 | _movstii : int -> int -> instr            (* movw n (n)   *)
 | _movstir : int -> int -> reg -> instr     (* movw n n'(r) *)
**********************************)


(** jump and return Instructions *)
Inductive sinstr : Set :=
   _jmp  : lab -> sinstr
 | _retw : sinstr
 | _iret : sinstr.

(** command, all the instructions and s-instructions are all commands. *)
Inductive comm : Set :=
  comml : instr  -> comm
 |comms : sinstr -> comm.

(** instruction sequence *)
Inductive iseq : Set :=
   seq  : instr -> iseq -> iseq
 | sseq : sinstr -> iseq.

(** CodeHeap C with Partitial Map **)
Module CHSpec <: MapSpec.
  Definition domain := lab.
  Definition image  := comm.
  Definition beq_Domain := eq_nat_dec.
End CHSpec.

(** CodeHeap C **)
Module CHMod := MapFun (CHSpec).
Definition CodeHeap := CHMod.Map.

(** instruction [l] code heap *)
Definition l_inC (ch : CodeHeap) (l : lab) (b : comm) : Prop :=
  CHMod.maps_to ch l b.

(** C[f]: find iseq in CodeHeap form label l, look up code heap **)
Inductive lookupC: CodeHeap -> lab -> iseq -> Prop :=
 | lookupC_c : forall (ch:CodeHeap) (l:lab) (s:sinstr),
           l_inC ch l (comms s)
         -> lookupC ch l (sseq s)
 | lookupC_b : forall (ch:CodeHeap) (l:lab) (s:instr) (b:iseq),
           l_inC ch l (comml s) 
         -> lookupC ch (l+1) b
         -> lookupC ch l (seq s b).

Definition World := (CodeHeap * State * PC)%type.

Definition W_PC (W : World) :=
  match W with
    (_, pc) => pc
  end.
  
Definition W_S (W : World) :=
  match W with
    (_, st, _) =>  st
  end.

Definition W_Ch (W : World) :=
  match W with
    (cdhp, _, _) => cdhp
  end.

Ltac UnfoldStateAll :=
  unfold S_Mem_Upd in * |- *; 
  unfold S_RF_Upd in * |- *; 
  unfold S_Flag_Upd in * |- *; 
  unfold S_ISR_Upd in * |- *;
  unfold S_Mem in * |- *; 
  unfold S_RF in * |- *; 
  unfold S_Flag in * |- *; 
  unfold S_ISR in * |- * .

(**  Iseq[l] in code block C1, it must in code block C1#C2 **)
Lemma lkpC_DisjUnion_Left:
  forall (C_1 C_2 C:CodeHeap) (x:iseq) (l:lab),
    lookupC C_1 l x /\ CHMod.disjUnion C_1 C_2 C
    -> lookupC C l x.
Proof.
induction x.
intros.
InvertAll.

inversion_clear H0.
generalize H1.
intros.
constructor.
 unfold CHMod.disjUnion in H1.
 InvertAll.
 unfold l_inC in H.
 unfold l_inC.
 rewrite H4.
 apply CHMod.in_left_in_merge; auto.

 apply IHx.
 split; auto.
intros.
InvertAll.
inversion_clear H0.
constructor.
unfold CHMod.disjUnion in H1.
InvertAll.
unfold l_inC in H.
unfold l_inC.
rewrite H2.
apply CHMod.in_left_in_merge; auto.
Qed.

(**  Iseq[l] in code block C2, it must in code block C1#C2 **)
Lemma lkpC_DisjUnion_Right:
  forall C_1 C_2 C l x,
    lookupC C_2 l x /\ CHMod.disjUnion C_1 C_2 C
    -> lookupC C l x.
Proof.
intros.
InvertAll.
apply lkpC_DisjUnion_Left with C_2 C_1.
split.
  auto.
  apply CHMod.disjUnion_commut.
  auto.
Qed.

(**  Iseq[l] in code block C is the only instruction sequnce **)
Lemma l_inC_unique:
  forall ch l b b',
    l_inC ch l b -> l_inC ch l b'
    -> b = b'.
Proof.
  intros.
  inversion H.
  inversion H0.
  rewrite H2 in H3.
  inversion H3.
  auto.
Qed.

(** lookup into code block C, can find only one unique Iseq[l]**)
Lemma lookupC_unique:
  forall (ch : CodeHeap) (b b' : iseq) (l : lab),
    lookupC ch l b -> lookupC ch l b'
    -> b = b'.
Proof.
induction b.
 induction b'.

  intros.
    inversion H.
    inversion H0.
    assert (comml i = comml i0).
   apply l_inC_unique with ch l; auto.
   inversion H13.
     assert (b = b').
    apply IHb with (l + 1); auto.
    rewrite H14; auto.

  intros.
    inversion H.
    inversion H0.
    assert (comml i = comms s).
   apply l_inC_unique with ch l; auto.
   inversion H11.

 intros.
   inversion H.
   inversion H0.
  assert (comms s = comms s1).
   apply l_inC_unique with ch l; auto.
   inversion H9; auto.
  assert (comms s = comml s1).
   apply l_inC_unique with ch l; auto.
   inversion H10; auto.
Qed.

(***added by wangwei***)
Lemma beq_reg_refl:
  forall r , true = beq_reg r r.
Proof.
induction r;simpl;trivial.
Qed.

Lemma reg_eq_dec:
  forall r r':reg,{r=r'}+{r<>r'}.
Proof.
double induction r r';
try (right;discriminate);try (left;reflexivity).
Qed.

Lemma reg_eq_beq_true:
  forall r r', r=r' -> true = beq_reg r r'.
Proof.
double induction r r';
try (intro H;elim H);
try apply beq_reg_refl.
Qed.

Lemma reg_neq_beq_false:
  forall r r', r<>r'->false = beq_reg r r'.
Proof.
double induction r r';
auto.
Qed.

Lemma reg_beq_true_eq:
  forall r r', true = beq_reg r r' -> r=r'.
Proof.
double induction r r';
simpl;intro H;
auto;inversion H.
Qed.

Lemma reg_beq_false_neq:
  forall r r',false = beq_reg r r'->r<>r'.
Proof.
double induction r r';
auto.
Qed.

Lemma beq_flag_refl:
  forall f , true = beq_flag f f.
Proof.
induction f;simpl;trivial.
Qed.

Lemma flag_eq_dec:
  forall r r':flag,{r=r'}+{r<>r'}.
Proof.
double induction r r';
try (right;discriminate);try (left;reflexivity).
Qed.

Lemma flag_eq_beq_true:
  forall r r', r=r' -> true = beq_reg r r'.
Proof.
double induction r r';
try (intro H;elim H);
try apply beq_reg_refl.
Qed.

Lemma flag_neq_beq_false:
  forall r r', r<>r'->false = beq_reg r r'.
Proof.
double induction r r';
auto.
Qed.

Lemma flag_beq_true_eq:
  forall r r', true = beq_reg r r' -> r=r'.
Proof.
double induction r r';
simpl;intro H;
auto;inversion H.
Qed.

Lemma falg_beq_false_neq:
  forall r r',false = beq_reg r r'->r<>r'.
Proof.
double induction r r';
auto.
Qed.

Lemma RF_update_RF_update : forall R r v v',
  RF_update (RF_update R r v') r v = RF_update R r v.
Proof.
intros.
apply (extensionality reg nat).
intro r0.
unfold RF_update.
generalize (beq_reg r0 r).
induction b;reflexivity.
Qed.

Lemma RF_update_RF : forall R r v,
  RF_update R r v r = v.
Proof.
unfold RF_update.
intros.
replace (beq_reg r r) with true.
reflexivity.
apply beq_reg_refl.
Qed.

Lemma RF_update_prev : forall R r,
  RF_update R r (R r) = R.
Proof.
intros.
apply (extensionality reg nat).
intro r0;unfold RF_update.
destruct (reg_eq_dec r0 r).
subst.
replace (beq_reg r r) with true;
[reflexivity|apply beq_reg_refl].
replace (beq_reg r0 r) with false;
[reflexivity|apply flag_neq_beq_false;assumption ].
Qed.
