(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Map                                                         *)
(*    using listT                                              *)
(*                                                             *)
(***************************************************************)

Require Import Coq.Bool.Bool.

Require Import Coqlib.
Require Export ListT.


Set Implicit Arguments.
Unset Strict Implicit.

Open Local Scope listt_scope.
Open Local Scope t_type_scope.

Module Type ONE_TYPE.

  Parameter B : Type.

End ONE_TYPE.

Module MultiMap (BDef : ONE_TYPE).

Definition A := nat.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := listT (prodT A B).

Definition emp_map : map := nilT.

Definition sig_map (a : A) (b : B) := (a, b) :: nilT.

Definition merge (m1 m2 : map) : map := 
  appT m1 m2.

Definition lookup (m1 : map) (a : A) (b : B) :=
  In (a, b) m1.

Definition indom (m : map) (a : A) :=
  exists b, lookup m a b.

Definition sub (m1 m2 : map) :=
  forall (a : A) (b : B), 
    lookup m1 a b -> lookup m2 a b.

Lemma sub_refl : forall m, sub m m.
Proof.
  unfold sub; auto.
Qed.

Lemma sub_trans : forall m1 m2 m3,
  sub m1 m2 -> sub m2 m3 -> sub m1 m3.
Proof.
  unfold sub, lookup; intros m1 m2 m3 H12 H23.
  intros a b.
  assert (H := (@incl_tran _ m1 m2 m3)).
  unfold incl in H.
  generalize (a, b).
  apply H.
  intros a0.
  destruct a0.
  apply H12; trivial.
  intros a0.
  destruct a0.
  apply H23; trivial.
Qed.

Lemma lookup_cons_intro :
  forall (m : map) (a a' : A) (b b' : B),
    lookup m a b
    -> lookup (consT (a', b') m ) a b.
Proof.
  unfold lookup; intros.
  simpl.
  right; trivial.
Qed.

Lemma lookup_merge_elim : 
  forall (m1 m2 : map) (a : A) (b : B),
    lookup (merge m1 m2) a b 
    -> lookup m1 a b \/ lookup m2 a b.
Proof.
  unfold lookup, merge.
  intros m1 m2 a b H.
  apply in_appT_or; trivial.
Qed.

Lemma lookup_merge_intro :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m1 a b \/ lookup m2 a b
    -> lookup (merge m1 m2) a b.
Proof.
  unfold lookup, merge; intros.
  apply in_or_appT; trivial.
Qed.

Lemma lookup_merge_intro_l :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m1 a b
    -> lookup (merge m1 m2) a b.
Proof.
  unfold lookup, merge; intros.
  apply in_or_appT; trivial.
  left; trivial.
Qed.

Lemma lookup_merge_intro_r :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m2 a b
    -> lookup (merge m1 m2) a b.
Proof.
  unfold lookup, merge; intros.
  apply in_or_appT; trivial.
  right; trivial.
Qed.

Lemma sub_merge_elim_l :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m1 m3.
Proof.
  unfold sub.
  intros m1 m2 m3 Hsub a b Hin1.
  apply Hsub.
  apply lookup_merge_intro.
  left; trivial.
Qed.
Implicit Arguments sub_merge_elim_l [m1 m2 m3].

Lemma sub_merge_elim_r :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m2 m3.
Proof.
  unfold sub.
  intros m1 m2 m3 Hsub a b Hin2.
  apply Hsub.
  apply lookup_merge_intro.
  right; trivial.
Qed.
Implicit Arguments sub_merge_elim_r [m1 m2 m3].

Lemma sub_merge_l :
  forall (m1 m2 : map),
    sub m1 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold sub, merge, lookup.
  intros a b H.
  apply in_or_appT.
  left; trivial.
Qed.

Lemma sub_merge_r :
  forall (m1 m2 : map),
    sub m2 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold sub, merge, lookup.
  intros a b H.
  apply in_or_appT.
  right; trivial.
Qed.

Lemma lookup_sub_lookup :
  forall (m1 m2 : map) (a : A) (b : B),
    lookup m1 a b 
    -> sub m1 m2 
    -> lookup m2 a b.
Proof.
  unfold sub; intros; auto.
Qed.

Lemma sub_merge_sub:
  forall (m1 m2 m3 : map),
    sub m1 m2
    -> sub (merge m3 m1) (merge m3 m2).
Proof.
  unfold sub.
  intros.
  assert (lookup m3 a b \/ lookup m1 a b).
  apply lookup_merge_elim; auto.
  apply lookup_merge_intro.
  inversion H1.
  left; auto.
  right; apply H; auto.
Qed.

Lemma lookup_sub_merge_l_lookup :
  forall (m1 m2 m3 : map) (a : A) (b : B),
    lookup m1 a b
    -> sub (merge m1 m2) m3
    -> lookup m3 a b.
Proof.
  intros m1 m2 m3 a b Hl Hsub.
  apply lookup_sub_lookup with m1; trivial.
  apply sub_merge_elim_l with m2; trivial.
Qed.
Implicit Arguments lookup_sub_merge_l_lookup [m1 m2 m3 a b].

Lemma lookup_sub_merge_r_lookup :
  forall (m1 m2 m3 : map) (a : A) (b : B),
    lookup m2 a b
    -> sub (merge m1 m2) m3
    -> lookup m3 a b.
Proof.
  intros m1 m2 m3 a b Hl Hsub.
  apply lookup_sub_lookup with m2; trivial.
  apply sub_merge_elim_r with m1; trivial.
Qed.
Implicit Arguments lookup_sub_merge_l_lookup [m1 m2 m3 a b].

Ltac simpl_In :=
  match goal with
    | |- In (pairT ?a ?b) (consT (pairT ?c ?d) ?m) =>
      unfold In; fold In; 
        (left; reflexivity; fail)
        || (right; simpl_In)
    | _ => fail 1 "fail"
  end.

Ltac lookup_solve :=
  match goal with
    | |- lookup ?m ?a ?b => unfold lookup; (try unfold m); simpl_In
    | _ => fail 1 
  end.

End MultiMap.

Unset Implicit Arguments.
