(****** Operational Semantics ******)

Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.

Set Implicit Arguments.
Unset Strict Implicit.

(* Definition stkSpace (sp : int) (n : nat) : MPred := *)
(*   fun (M : Mem) => (sp >= n) /\ (leadsToBlk (sp-n) n) M. *)

(* (\* precise version of stkSpace *\) *)
(* Definition stkSpacePrec (sp : int) (n : nat) : MPred := *)
(*   fun (M : Mem) => (sp >= n) /\ (ptoxBlk (sp-n) n) M. *)


(* Definition MemUpd16 (M : Mem) (l : int) (n : int) (M' : Mem) : Prop := *)
(*   exists x, M' = MemMod.put (MemMod.put M l n) (l+1) x. *)

Definition opr_eval (R : RFile) (o : Opr) : int :=
  match o with
    Opr_imm i => i
  | Opr_reg r => R r
  end.

Definition addr_eval (R : RFile) (a : Addr) : int :=
  match a with
    A_imm l => l
  | A_reg l r => R r + l
  end.

Definition pushW (M M' : Mem) (R R': RFile) (n : int) : Prop :=
  stkSpace (R SP) 2 M /\ MemUpd16 M (R SP - 2) n M' /\ R' = RF_update R SP (R SP - 2). 

Definition popW (M : Mem) (R R': RFile) (n : int) : Prop :=
  leadsToW (R SP) n M /\ R' = RF_update R SP (R SP + 2).

Definition trash_fl (FL FL': FlagR) (f : flag) : Prop :=
  (FL' = Set_Flag FL f) \/ (FL' = Clear_Flag FL f).

Inductive nextS : PC -> comm -> State -> State -> Prop :=
  (** Interrupt status instructions **)
  | nextS_cli:
     forall (pc : PC) (M:Mem) (R:RFile) (fl: FlagR) (isr: ISR),
           nextS pc (comml _cli) (M, R, fl, isr) (M, R, Clear_Flag fl _IF, isr)
  | nextS_sti:
     forall (pc : PC) (M:Mem) (R:RFile) (fl: FlagR) (isr: ISR),
           nextS pc (comml _sti) (M, R, fl, isr) (M, R, Set_Flag fl _IF, isr)
  (** Function call instruction**)
  | nextS_call:
     forall (pc : PC) (f: flab) (M M':Mem) (R R':RFile) (fl : FlagR) (isr: ISR),
       pushW M M' R R' (pc + 1)
       -> nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl, isr)
  (** BrancM instruction **)
  | nextS_jne :
     forall (pc : PC) (f : flab) (M:Mem) (R: RFile) (fl : FlagR) (isr:ISR),
       nextS pc (comml (_jne f)) (M, R, fl, isr) (M, R, fl, isr)
  (** Move instruction **)
  | nextS_movwp :
     forall (pc : PC) (o : Opr) (r : reg) (M:Mem) (R R': RFile) (fl : FlagR) (isr: ISR) (x : int),
       (x = opr_eval R o)
       -> R' = RF_update R r x
       -> nextS pc (comml (_movwp o r)) (M, R, fl, isr) (M, R', fl, isr)
  | nextS_movwld :
     forall (pc : PC) (r : reg) (a : Addr) (M:Mem) (R R': RFile) (fl : FlagR) (isr: ISR) (l x : int),
       (l = addr_eval R a)
       -> leadsToW l x M
       -> R' = RF_update R r x
       -> nextS pc (comml (_movwld a r)) (M, R, fl, isr) (M, R', fl, isr)
  | nextS_movwst :
     forall (pc : PC) (o : Opr) (a : Addr) (M M':Mem) (R: RFile) (fl : FlagR) (isr: ISR) (l x : int),
       (l = addr_eval R a)
       -> (x = opr_eval R o)
       -> leadsToxW l M
       -> MemUpd16 M l x M'
       -> nextS pc (comml (_movwst o a)) (M, R, fl, isr) (M', R, fl, isr)
  (** stack operations **)
  | nextS_pushwp :
     forall (pc : PC) (o : Opr) (M M':Mem) (R R': RFile) (fl : FlagR) (isr: ISR) (x:int),
       (x = opr_eval R o)
       -> pushW M M' R R' x
       -> nextS pc (comml (_pushwp o)) (M, R, fl, isr) (M', R', fl, isr)
  | nextS_pushwm :
     forall (pc : PC) (a : Addr) (M M':Mem) (R R': RFile) (fl : FlagR) (isr: ISR) (l x :int),
       (l = addr_eval R a)
       -> leadsToW l x M
       -> pushW M M' R R' x
       -> nextS pc (comml (_pushwm a)) (M, R, fl, isr) (M', R', fl, isr)
  | nextS_pushfw :
     forall (pc : PC) (M M':Mem) (R R': RFile) (fl : FlagR) (isr: ISR),
       pushW M M' R R' (FR2nat fl)
       -> nextS pc (comml _pushfw) (M, R, fl, isr) (M', R', fl, isr)
  | nextS_popw :
     forall (pc : PC) (r : reg) (M:Mem) (R R' R'': RFile) (fl : FlagR) (isr: ISR) (x:int),
       popW M R R' x
       -> R'' = RF_update R' r x
       -> nextS pc (comml (_popw r)) (M, R, fl, isr) (M, R'', fl, isr)
  | nextS_popfw :
     forall (pc : PC) (M:Mem) (R R': RFile) (fl fl': FlagR) (isr: ISR) (flnum:int),
       popW M R R' flnum
       -> fl' = nat2FR flnum
       -> nextS pc (comml _popfw) (M, R, fl, isr) (M, R', fl', isr)

  (** Simple instructions **)
  | nextS_cmpw :
     forall (pc : PC) (o : Opr) (r : reg) (M:Mem) (R: RFile) (fl fl': FlagR) (isr:ISR),
       ((opr_eval R o = R r) /\ (fl' = Set_Flag fl _ZF)
          \/ (opr_eval R o <> R r) /\ (fl' = Clear_Flag fl _ZF))
       -> nextS pc (comml (_cmpw o r)) (M, R, fl, isr) (M, R, fl', isr)
  | nextS_addw :
     forall (pc : PC) (o : Opr) (r : reg) (M : Mem) (R R' : RFile) (fl fl' : FlagR) (isr : ISR) (x x' : int),
       (x = opr_eval R o)
       -> (x' = R r)
       -> R' = RF_update R r (x + x')
       -> trash_fl fl fl' _ZF
       -> nextS pc (comml (_addw o r)) (M, R, fl, isr) (M, R', fl', isr)
  | nextS_subw :
     forall (pc : PC) (o : Opr) (r : reg) (M : Mem) (R R' : RFile) (fl fl' : FlagR) (isr : ISR) (x x' : int),
       (x = opr_eval R o)
       -> (x' = R r)
       -> R' = RF_update R r (x' - x)
       -> trash_fl fl fl' _ZF
       -> nextS pc (comml (_subw o r)) (M, R, fl, isr) (M, R', fl', isr)
  (** eoi instruction **)
  | nextS_EOI :
     forall (pc : PC) (M : Mem) (R : RFile) (fl : FlagR) (isr : ISR),
       (R AX = 32)
       -> (R DX = 32)
       -> nextS pc (comml _outb_eoi) (M, R, fl, isr) (M, R, fl, false)
  
  (** Special instruction **)
  | nextS_jmp :
     forall (pc : PC) (M:Mem) (R: RFile) (fl : FlagR) (isr:ISR) (f': flab),
           nextS pc (comms (_jmp f')) (M, R, fl, isr) (M, R, fl, isr)
  | nextS_retw:
     forall (pc : PC) (M:Mem) (R R':RFile) (fl : FlagR) (isr: ISR),
       (exists pc', popW M R R' pc')
       -> nextS pc (comms _retw) (M, R, fl, isr) (M, R', fl, isr)
  | nextS_iret:
     forall (pc : PC) (M:Mem) (R R':RFile) (fl fl': FlagR) (isr: ISR) (flnum sp : int),
       (R SP = sp) 
       -> leadsToBlk sp 6 M 
       -> leadsToW (sp+4) flnum M 
       -> fl' = nat2FR flnum
       -> R' = RF_update R SP (sp+6)
       -> nextS pc (comms _iret) (M, R, fl, isr) (M, R', fl', isr).

Inductive nextPC : comm -> Mem -> RFile -> FlagR -> PC -> PC -> Prop :=
  (** Interrupt status instructions **)
  | nextPC_cli:
     forall (pc : PC) (M:Mem) (R:RFile) (fl:FlagR),
           nextPC (comml _cli) M R fl pc (pc+1)
  | nextPC_sti:
     forall (pc : PC) (M:Mem) (R:RFile) (fl:FlagR),
           nextPC (comml _sti) M R fl pc (pc+1)
  (** Function call instruction**)
  | nextPC_call:
     forall (pc : PC) (f: flab) (M :Mem) (R :RFile) (fl:FlagR),
           nextPC (comml (_call f)) M R fl pc f
  (** BrancM instruction **)
  | nextPC_jneF:
     forall (pc : PC) (f : flab) (M:Mem) (R: RFile) (fl : FlagR),
       (fl _ZF = true)
       -> nextPC (comml (_jne f)) M R fl pc (pc+1)
  | nextPC_jneT:
     forall (pc : PC) (f : flab) (M:Mem) (R: RFile) (fl : FlagR),
       (fl _ZF = false)
       -> nextPC (comml (_jne f)) M R fl pc f
  (** Move instruction **)
  | nextPC_movwp :
     forall (pc : PC) (o : Opr)  (r : reg) (M:Mem) (R : RFile) (fl:FlagR),
       nextPC (comml (_movwp o r)) M R fl pc (pc+1)
  | nextPC_movwld :
     forall (pc : PC) (a : Addr) (r : reg) (M:Mem) (R : RFile) (fl:FlagR),
       nextPC (comml (_movwld a r)) M R fl pc (pc+1)
  | nextPC_movwst :
     forall (pc : PC) (o : Opr) (a : Addr) (M:Mem) (R : RFile) (fl:FlagR),
       nextPC (comml (_movwst o a)) M R fl pc (pc+1)
  (** stack operations **)
  | nextPC_pushwp :
     forall (pc : PC) (o : Opr)  (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_pushwp o)) M R fl pc (pc+1)
  | nextPC_pushwm :
     forall (pc : PC) (a : Addr) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_pushwm a)) M R fl pc (pc+1)
  | nextPC_pushfw :
     forall (pc : PC) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml _pushfw) M R fl pc (pc+1)
  | nextPC_popw :
     forall (pc : PC) (r : reg) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_popw r)) M R fl pc (pc+1)
  | nextPC_popfw :
     forall (pc : PC) (M:Mem) (R : RFile) (fl:FlagR),
       nextPC (comml _popfw) M R fl pc (pc+1)
  (** Simple instructions **)
  | nextPC_cmpw :
     forall (pc : PC) (o : Opr) (r : reg) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_cmpw o r)) M R fl pc (pc+1)
  | nextPC_addw :
     forall (pc : PC) (o : Opr) (r : reg) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_addw o r)) M R fl pc (pc+1)
  | nextPC_subw :
     forall (pc : PC) (o : Opr) (r : reg) (M:Mem) (R: RFile) (fl:FlagR),
       nextPC (comml (_subw o r)) M R fl pc (pc+1)
  (** eoi instruction **)
  | nextPC_EOI :
     forall (pc : PC) (M : Mem) (R : RFile) (fl:FlagR),
       nextPC (comml _outb_eoi) M R fl pc (pc+1)
  (** Special instruction **)
  | nextPC_jmp :
     forall (pc : PC) (M:Mem) (R: RFile) (f': flab) (fl:FlagR),
       nextPC (comms (_jmp f')) M R fl pc f'
  | nextPC_retw:
     forall (pc pc': PC) (M:Mem) (R:RFile) (sp:int) (fl:FlagR),
       (R SP = sp)
       -> leadsToW sp pc' M
       -> nextPC (comms _retw) M R fl pc pc'
  | nextPC_iret:
     forall (pc pc': PC) (M:Mem) (R:RFile) (sp:int) (fl:FlagR),
       (R SP = sp)
       -> leadsToBlk sp 6 M
       -> leadsToW sp pc' M
       -> nextPC (comms _iret) M R fl pc pc'.


(** Step World:  W --> W' common instructions **)
Definition stepInstr (W:World) (W':World) : Prop :=
  match W,W' with
    (ch, st, pc), (ch',st',pc')=>
                  (ch = ch') 
                  /\ exists c:comm, l_inC ch pc c 
                    /\ nextPC c (S_Mem st) (S_RF st) (S_Flag st) pc pc'
                    /\ nextS pc c st st'
  end.

Definition itr_enabled (S : State) : Prop :=
  (S_Flag S _IF = true) /\ (S_ISR S = false).

Definition itr_nextS (pc : PC) (S S' : State) : Prop :=
  match S, S' with
    | (M, R, fl, isr), (M',R',fl',isr')=>
      exists sp, exists flnum, exists cs, exists M'', exists M''', 
        (sp = R SP)
        /\ (stkSpace sp 6 M) 
        /\ (R' = RF_update R SP (sp - 6)) 
        /\ (fl' = Clear_Flag fl _IF) 
        /\ (isr' = true) 
        /\ (flnum = FR2nat fl)
        /\ (MemUpd16 M (sp-6) pc M'') 
        /\ (MemUpd16 M'' (sp-4) cs M''') 
        /\ (MemUpd16 M''' (sp-2) flnum M')
  end.
                
(** Step World: W ==> W' interrupts *)
Definition stepItr  (W:World) (W':World) : Prop :=
  match W, W' with
    | (ch, st, pc), (ch', st', pc') =>
            (ch = ch') 
            /\ itr_enabled st 
            /\ itr_nextS pc st st' 
            /\ pc' = h_entry
  end.

(** Step World ****)
Inductive step : World -> World -> Prop :=
  (** W --> W' common instructions **)
  | step_instr :
     forall (W:World) (W':World),
       stepInstr W W'
       -> step W W'
  (** W ==> W' interrupts *)
  | step_inter :
     forall (W:World) (W':World),
       stepItr W W'
       -> step W W'.

(** N-Step World ****)
Inductive stepN : nat -> World -> World -> Prop :=
  (** 0-Step: 0 W |==> W **)
  | stepO :
     forall W W',
       W' = W
       -> stepN 0 W W'
  (** N-Step: n W |==> W'**)
  | stepS :
     forall W W' n,
       (exists W'', step W W'' /\ stepN n W'' W')
       -> stepN (S n) W W'.


