Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.

Require Import EqNat.

Definition TID := nat.

(**************************************************)
(* Definitions of TCBs                            *)
(**************************************************)

(* thread status *)
Inductive TStatus : Set :=
    IDLE      (* 0 *)
  | READY     (* 1 *)
  | BLOCKED.  (* 2 *)


(* abstract TCB : stack pointer, status, return code pointer *)
Definition Thread := (Label * TStatus * PC)%type.
(* Definition Thread := (Label * TStatus)%type. *)

Definition TCB_SP (tcb : Thread) : Label :=
  match tcb with
    (sp, _, _) => sp
  end.

Definition TCB_PC (tcb : Thread) : PC :=
  match tcb with
    (_, pc) => pc
  end.

Definition TCB_Status (tcb : Thread) : TStatus :=
  match tcb with
    (_, status, _) => status
  end.

Definition TCB_SP_Upd (tcb : Thread) (sp : Label) : Thread :=
  match tcb with
    (sp0, tst, pc) => (sp, tst, pc)
  end.

Definition TCB_PC_Upd (tcb : Thread) (pc : PC) : Thread :=
  match tcb with
    (sp, tst, pc0) => (sp, tst, pc)
  end.

Definition TCB_Status_Upd (tcb : Thread) (tst : TStatus) : Thread :=
  match tcb with
    (sp, tst0, pc) => (sp, tst, pc)
  end.

Definition nat2Status (n : nat) : option TStatus :=
  match n with
    0 => Some IDLE
  | S n' =>
    match n' with
      0 => Some READY
    | S n'' =>
      match n'' with
        0 => Some BLOCKED
      | _ => None
      end
    end
  end.

Definition Status2nat (ts : TStatus) : nat :=
  match ts with
    | IDLE => 0
    | READY => 1
    | BLOCKED => 2
  end.



(* below we embed the abstract TCBs into Memory          *)
(* Note that this embedding does not save PC in TCB      *)
Definition ptoTCB (l : Label) (tcb : Thread) : MPred :=
  fun (M : Mem) =>
    match tcb with
      (sp, status, pc) =>
          exists w, mstar (ptoW l sp) (ptoW (l+2) w) M
                    /\ nat2Status w = Some status
    end.


(* Below we define the relationship between PC and SP in TCB *)
Definition TCB_SP2PC (tcb : Thread) : MPred :=
  (TCB_SP tcb) |->w (TCB_PC tcb).


(********************************************************)
(*  Definitions of thread queues                        *)
(********************************************************)
(** ThreadQ T as a Partitial Map **)
Module TQSpec <: MapSpec.
  Definition domain := TID.
  Definition image  := Thread.
  Definition beq_Domain := eq_nat_dec.
End TQSpec.

Module TQMod := MapFun (TQSpec).
Definition TQ := TQMod.Map.


(********** TQ in memory  ***************)
Definition ptoTQ := ptoQ Thread ptoTCB.

(*
Inductive embed_PCinTQ : TQ -> Mem -> Prop :=
    PCinTQ_emp  : embed_PCinTQ TQMod.emp MemMod.emp
  | PCinTQ_cons : forall T M tid tcb,
                    ((embed_PCinTQ T) *^* (TCB_SP2PC tcb)) M
                    -> TQMod.not_in_dom tid T
                    -> embed_PCinTQ (TQMod.put T tid tcb) M.
*)

(** We use a partial mapping to DummyImg to implement sets **)
Inductive DummyImg : Set :=
  dummy_sglton : DummyImg.

Module WQSpec <: MapSpec.
  Definition domain := TID.
  Definition image := DummyImg.
  Definition beq_Domain := eq_nat_dec.
End WQSpec.

Module WQMod := MapFun WQSpec.
Definition WQ := WQMod.Map.

Definition addThrd (Q : WQ) (tid : TID) : WQ :=
  WQMod.put Q tid dummy_sglton.

Definition filterTQ (T : TQ) (Q : WQ) : TQ :=
  fun (tid : TID) =>
    match Q tid with
      Some _ => T tid
    | None   => None
    end.

Definition subsetQ_TQ (Q : WQ) (T : TQ) : Prop :=
  forall tid, WQMod.in_dom tid Q -> TQMod.in_dom tid T.

Module WQSetSpec <: MapSpec.
  Definition domain := Label.
  Definition image := WQ.
  Definition beq_Domain := eq_nat_dec.
End WQSetSpec.

Module WQSetMod := MapFun WQSetSpec.
Definition WQSet := WQSetMod.Map.

(* ThrdMem is a partial mapping from Thread IDs to heaps *)
(* It records the partition of memory among threads      *)
Module ThrdMemSpec <: MapSpec.
  Definition domain := TID.
  Definition image := Mem.
  Definition beq_Domain := eq_nat_dec.
End ThrdMemSpec.

Module ThrdMemMod :=  MapFun(ThrdMemSpec).

Definition ThrdMem := ThrdMemMod.Map.

Inductive MemSplit :  ThrdMem -> Mem -> Prop :=
    MS_empty: MemSplit ThrdMemMod.emp MemMod.emp
  | MS_split: forall (M M' M'' : Mem) (TM : ThrdMem) (tid : TID),
                MemMod.disjUnion M' M'' M ->
                MemSplit TM M' ->
                ThrdMemMod.not_in_dom tid TM ->
                MemSplit (ThrdMemMod.put TM tid M'') M.

(****************************************************)
(*  Interpretations for AIM logics                  *)
(****************************************************)
Definition NH_Inv (fl : FlagR) : MPred :=
  if (fl _IF) then mstar INV0 INV1 else Aemp.

Lemma NHInv_prec : forall fl, precise (NH_Inv fl).
Proof.
  intro.
  unfold NH_Inv.
  induction (fl _IF).
  apply mstar_presv_prec; auto.
  apply Prec0. apply Prec1.
  apply precise_emp.
Qed.

(** p*Inv   **)
Definition NH_pinvstar (p : Pred) : Pred :=
  fun (S : State) =>
    pmstar p (NH_Inv (S_Flag S)) S.

Definition NH_ItpG (g : Grt) : Grt :=
  fun (S S' : State) =>
    exists M1, exists M2, exists M1', exists M2',
      MemMod.disjUnion M1 M2 (S_Mem S) /\ MemMod.disjUnion M1' M2' (S_Mem S')
        /\ NH_Inv (S_Flag S)  M2
	/\ NH_Inv (S_Flag S') M2'
        /\ g (S_Mem_Upd S M1) (S_Mem_Upd S' M1').


Definition rho_ready : LangID := 1.
Definition rho_idle  : LangID := 2.
Definition rho_ih    : LangID := 3.
Definition rho_prim  : LangID := 4.


Fixpoint NH_WFST_aux (rho : LangID) (n : nat) (Psi : PSpec) (p : Pred) {struct n} : Prop :=
  match n with
    0 => forall S, p S -> exists ra, Ret_Addr ra S
                           /\ exists p', exists g', exists tag,
                                 lookupCT Psi ra (packS0 rho (Ast p' g' tag))
                                 /\ tag <> 1
                                 /\ forall pc S', (nextS pc (comms _retw) S S') ->  NH_pinvstar p' S'
  | S n' =>
         forall S, p S -> exists ra, Ret_Addr ra S
                          /\ exists p', exists g', exists tag,
                                 lookupCT Psi ra (packS0 rho (Ast p' g' tag))
                                 /\ forall pc S', 
                                    nextS pc (comms _retw) S S' ->
                                        NH_pinvstar p' S' /\ NH_WFST_aux rho n' Psi (NH_ItpG g' S')
  end.


Definition NH_WFST (rho : LangID) (Psi : PSpec) (p : Pred) : Prop :=
  exists n, NH_WFST_aux rho n Psi p.

(* Definition NH_WFST2 (Psi : PSpec) (g : Grt) : Pred := *)
(*   fun (S : State) => NH_WFST Psi (g S). *)

(* a aux interpretation of NH spec *)
Definition itp_NH_theta (rho : LangID) (theta : Spec) (Psi : PSpec) (S: State) : Prop :=
  S_ISR S = false /\
  match theta with
    Ast p g tag =>
      NH_pinvstar p S /\ (tag = 1 -> NH_WFST rho Psi (NH_ItpG g S))
  end.

Definition IH_ItpG (g : Grt) : Grt :=
  fun (S S' : State) =>
    exists M1, exists M2, exists M1', exists M2',
      MemMod.disjUnion M1 M2 (S_Mem S) /\ MemMod.disjUnion M1' M2' (S_Mem S')
        /\ INV1 M2
	/\ INV1 M2'
        /\ g (S_Mem_Upd S M1) (S_Mem_Upd S' M1').

Definition rho_tcb_consistent2 (status : TStatus) (rho: LangID): Prop :=
 (rho = rho_ready -> status = READY) /\ (rho = rho_idle -> status = IDLE).

Definition lift_IH_PG (theta : Spec) (Psi : PSpec) (status: TStatus) (S : State) : Prop :=
 match theta with
   Ast p g tag =>
       tag = 0
       /\ pmstar p INV1 S
       /\ forall S', IH_ItpG g S S' ->
              exists S'', (forall pc', nextS pc' (comms _iret) S' S'')
              /\ exists ra, Ret_Addr ra S'
                            /\ exists rho, exists theta,
                                  (rho = rho_ready \/ rho = rho_idle)
                                  /\ rho_tcb_consistent2 status rho       (* this is a new line *)
                                  /\ lookupCT Psi ra (packS0 rho theta)
                                  /\ forall pc' S'', nextS pc' (comms _iret) S' S''
                                                     -> itp_NH_theta rho theta Psi S''
 end.

(* Definition lift_IH_PG (theta : Spec) (Psi : PSpec) (S : State) : Prop := *)
(*   match theta with *)
(*     Ast p g tag => *)
(*         tag = 0 *)
(*         /\ pmstar p INV1 S *)
(*         /\ forall S', IH_ItpG g S S' ->  *)
(*                exists S'', (forall pc', nextS pc' (comms _iret) S' S'') *)
(*                /\ exists ra, Ret_Addr ra S' *)
(*                              /\ exists rho, exists theta, *)
(*                                    (rho = rho_ready \/ rho = rho_idle) *)
(*                                    /\ lookupCT Psi ra (packS0 rho theta) *)
(*                                    /\ forall pc' S'', nextS pc' (comms _iret) S' S'' *)
(*                                                       -> itp_NH_theta rho theta Psi S'' *)
(*   end. *)


Fixpoint IH_WFST (n : nat) (Psi : PSpec) (p : Pred) (status: TStatus) {struct n} : Prop :=
 match n with
   0 => forall S, p S ->
          exists ra, Ret_Addr ra S
                     /\ exists theta,
                        lookupCT Psi ra (packS0 rho_ih theta)
                        /\ forall pc' S', nextS pc' (comms _retw) S S'
                                          -> lift_IH_PG theta Psi status S'
 | S n' =>
        forall S, p S ->
          exists ra, Ret_Addr ra S
                     /\ exists p', exists g', exists tag',
                        lookupCT Psi ra (packS0 rho_ih (Ast p' g' tag'))
                        (* modified by guoyu, every tag in stack should all
                         be non-top (1), so we may add a constraint below,
                        which is indispensable for the soundness proof. *)
                        /\ tag' = 1
                        /\ forall pc' S', nextS pc' (comms _retw) S S'
                                          -> pmstar p' INV1 S' /\ IH_WFST n' Psi (IH_ItpG g' S') status
 end.

(* Fixpoint IH_WFST (n : nat) (Psi : PSpec) (p : Pred) {struct n} : Prop := *)
(*   match n with *)
(*     0 => forall S, p S -> *)
(*            exists ra, Ret_Addr ra S  *)
(*                       /\ exists theta, *)
(*                          lookupCT Psi ra (packS0 rho_ih theta) *)
(*                          /\ forall pc' S', nextS pc' (comms _retw) S S' *)
(*                                            -> lift_IH_PG theta Psi S' *)
(*   | S n' => *)
(*          forall S, p S -> *)
(*            exists ra, Ret_Addr ra S  *)
(*                       /\ exists p', exists g', exists tag', *)
(*                          lookupCT Psi ra (packS0 rho_ih (Ast p' g' tag')) *)
(*                          (* modified by guoyu, every tag in stack should all  *)
(*                           be non-top (1), so we may add a constraint below,  *)
(*                          which is indispensable for the soundness proof. *) *)
(*                          /\ tag' = 1 *)
(*                          /\ forall pc' S', nextS pc' (comms _retw) S S' *)
(*                                            -> pmstar p' INV1 S' /\ IH_WFST n' Psi (IH_ItpG g' S') *)
(*   end. *)


(* a aux interpretation of IH spec *)

Definition itp_IH_theta (theta : Spec) (status : TStatus) (Psi : PSpec) (S : State) : Prop :=
 S_Flag S _IF = false /\
 match theta with
   Ast p g tag =>
     (lift_IH_PG theta Psi status S) \/ (tag = 1) /\ (pmstar p INV1) S /\ exists n, IH_WFST n Psi (IH_ItpG g S) status
 end.

(* Definition itp_IH_theta (theta : Spec) (Psi : PSpec) (S: State) : Prop := *)
(*   S_Flag S _IF = false /\ *)
(*   match theta with *)
(*     Ast p g tag => *)
(*       (lift_IH_PG theta Psi S) \/ (tag = 1) /\ (pmstar p INV1) S /\ exists n, IH_WFST n Psi (IH_ItpG g S) *)
(*   end. *)

(*
Definition WF_Rdy_Thrd (Psi : PSpec) (M : Mem) (tcb : Thread) :=
  match tcb with
    (X, pc) =>
    match X with
      (sp, status) =>
          exists rho, exists p, exists g, exists tag,
                (rho = rho_ready \/ rho = rho_idle)
                /\ leadsToW sp pc M
                /\ lookupCT Psi pc (packS0 rho (Ast p g tag)) 
                /\ forall R fl,
                      R SP = sp + 2
                      -> fl _IF = false 
                      -> (pm_mwand (mstar INV0 INV1) (itp_NH_theta rho (Ast p g tag) Psi)) (M, R, fl, false)
    end
  end.
*)

Definition rho_tcb_consistent (tcb : Thread) (rho: LangID): Prop :=
  (rho = rho_ready -> TCB_Status tcb = READY) /\ (rho = rho_idle -> TCB_Status tcb = IDLE).


Definition WF_Rdy_Thrd (Psi : PSpec) (M : Mem) (tcb : Thread) :=
  match tcb with
    (sp, status, pc) =>
    exists rho, exists p, exists g, exists tag,
      (rho = rho_ready \/ rho = rho_idle)
      /\ rho_tcb_consistent tcb rho
      /\ leadsToW sp pc M
      /\ lookupCT Psi pc (packS0 rho (Ast p g tag)) 
      /\ forall R fl,
        R SP = sp + 2
        -> fl _IF = false 
        -> (pm_mwand (mstar INV0 INV1) (itp_NH_theta rho (Ast p g tag) Psi)) (M, R, fl, false)
  end.

Definition WF_Rdy_Thrd_IH (Psi : PSpec) (M : Mem) (tcb : Thread) :=
 match tcb with
   (sp, status, pc) =>
   exists p, exists g, exists tag,
     lookupCT Psi pc (packS0 rho_ih (Ast p g tag))
     /\ leadsToW sp pc M
     /\ forall R fl,
       R SP = sp + 2
       -> fl _IF = false
       -> (pm_mwand (mstar INV0 INV1) (itp_IH_theta (Ast p g tag) status Psi )) (M, R, fl, false)
 end.

(* Definition WF_Rdy_Thrd_IH (Psi : PSpec) (M : Mem) (tcb : Thread) := *)
(*   match tcb with *)
(*     (sp, status, pc) => *)
(*     exists p, exists g, exists tag, *)
(*       lookupCT Psi pc (packS0 rho_ih (Ast p g tag)) *)
(*       /\ leadsToW sp pc M *)
(*       /\ forall R fl, *)
(*         R SP = sp + 2 *)
(*         -> fl _IF = false *)
(*         -> (pm_mwand (mstar INV0 INV1) (itp_IH_theta (Ast p g tag) Psi) ) (M, R, fl, false) *)
(*   end. *)

Definition WF_Curr_Thrd (Psi : PSpec) (S : State) (rho : LangID) (theta : Spec) (T: TQ) (tid : TID) :=
 if beq_nat rho rho_ih
 then exists tcb, exists status,
        TQMod.maps_to T tid tcb /\ TCB_Status tcb = status /\ itp_IH_theta theta status Psi S
   else itp_NH_theta rho theta Psi S.

(* Definition WF_Curr_Thrd (Psi : PSpec) (S : State) (rho : LangID) (theta : Spec) := *)
(*   if beq_nat rho rho_ih then itp_IH_theta theta Psi S *)
(*     else itp_NH_theta rho theta Psi S. *)

Definition WF_Wait_Thrd (Psi : PSpec) (M : Mem) (tcb : Thread) (l : Label) :=
  match tcb with
    (sp, status, pc) =>
    status = BLOCKED /\
    leadsToW sp pc M /\
    exists wqa, Delta l = SomeT wqa
      /\ exists p, exists g, exists tag,
        lookupCT Psi pc (packS0 rho_ready (Ast p g tag))
        (* /\ leadsToW sp pc M *)
        /\ forall R fl,
          R SP = sp + 2
          -> fl _IF = false
          -> (pm_mwand (mstar (mstar INV0 INV1) wqa) (itp_NH_theta rho_ready (Ast p g tag) Psi)) (M, R, fl, false)
  end.

Definition WF_WQ_WQSpec (B : WQSet) (Delta : WQSpec) : Prop :=
  forall (l : Label) (wqa : WQassert),
    Delta l = SomeT wqa -> WQSetMod.in_dom l B.


Definition WF_WQSet (Psi : PSpec) (T : TQ) (B : WQSet) (TM : ThrdMem) : Prop :=
  forall (l : Label) (Q : WQ) (tid : TID),
    WQSetMod.maps_to B l Q ->
    WQMod.in_dom tid Q ->
    exists tcb : Thread, exists M : Mem,
      TQMod.maps_to T tid tcb /\ ThrdMemMod.maps_to TM tid M /\
      WF_Wait_Thrd Psi M tcb l.

Definition unblocked (tid: TID) (B: WQSet) : Prop :=
  forall (w : Label) (Q : WQ), 
    WQSetMod.maps_to B w Q -> WQMod.not_in_dom tid Q.

(* last update by Xinyu : Dec. 8, 07 *)
Definition WF_ReadyQ (Psi : PSpec) (T : TQ) (RdyQ : WQ) (TM : ThrdMem) : Prop :=
  forall (tid : TID),
    WQMod.in_dom tid RdyQ -> 
      exists tcb : Thread, 
        TQMod.maps_to T tid tcb /\
        (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\ 
        exists M : Mem, ThrdMemMod.maps_to TM tid M /\
                         leadsToW (TCB_SP tcb) (TCB_PC tcb) M /\ 
                         (WF_Rdy_Thrd Psi M tcb \/ WF_Rdy_Thrd_IH Psi M tcb).


(*
Definition WF_currTID (tid : TID) (T : TQ) (B : WQSet) (RdyQ : WQ) : Prop :=
  unblocked tid B 
  /\ WQMod.not_in_dom tid RdyQ 
  /\ TQMod.in_dom tid T
  /\ exists tcb, TQMod.maps_to T tid tcb /\ (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE).
*)

Definition WF_currTID (tid : TID) (T : TQ) (B : WQSet) (RdyQ : WQ) (rho: LangID) : Prop :=
  unblocked tid B 
  /\ WQMod.not_in_dom tid RdyQ 
  /\ TQMod.in_dom tid T
  /\ exists tcb, TQMod.maps_to T tid tcb 
    /\ (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE)
    /\ rho_tcb_consistent tcb rho.


Definition WF_TQ (T : TQ) : Prop :=
  forall tid, TQMod.in_dom tid T -> tid <> 0.

Definition WF_WQSet2 (B : WQSet) : Prop :=
  forall l l' Q Q', WQSetMod.maps_to B l Q -> WQSetMod.maps_to B l' Q' -> l <> l' ->
    WQMod.disjoint Q Q'.

Definition disj_RQ_WQ (B : WQSet) (RdyQ : WQ) : Prop :=
  forall l Q, WQSetMod.maps_to B l Q -> WQMod.disjoint Q RdyQ.

Definition WF_TWQ (T : TQ) (B : WQSet) (RdyQ : WQ) : Prop :=
  WF_TQ T /\ WF_WQSet2 B /\ disj_RQ_WQ B RdyQ.

(** Specification of the Interrupte Handler **)
Definition IHpred : Pred :=
  fun (S : State) =>
    pmstar pTrue (mstar INV0 (stkSpacePrec (S_RF S SP + 6) (IH_Stk + 6))) S
      /\ (S_ISR S) = true
      /\ (S_Flag S _IF) = false.

Definition IHgrt : Grt :=
  fun (S S' : State) =>
    prepost (mstar INV0 (stkSpacePrec (S_RF S SP) IH_Stk)) 
            (mstar INV0 (stkSpacePrec (S_RF S' SP) IH_Stk)) S S'
    /\ S_RF S = S_RF S'
    /\ S_ISR S' = false.

Definition IH_Spec0 := (packS0 rho_ih (Ast IHpred IHgrt 0)).

Definition IH_Specified (Psi : PSpec) : Prop :=  lookupCT Psi h_entry IH_Spec0.


Definition exists_Idle (T : TQ) (Rdy : WQ) (tid : TID) : Prop :=
  (exists tcb, TQMod.maps_to T tid tcb /\ TCB_Status tcb = IDLE)
  \/ (exists tid', exists tcb, WQMod.in_dom tid' Rdy /\ TQMod.maps_to T tid' tcb /\ TCB_Status tcb = IDLE).

(* This is the global invariant enforced in AIM *)
Definition WF_AIM_World (Psi : PSpec) (S : State) (T : TQ) (B : WQSet) (RdyQ : WQ) (tid : TID) (langid : LangID) (theta : Spec) : Prop :=
     WF_currTID tid T B RdyQ langid
  /\ exists_Idle T RdyQ tid
  /\ WF_TWQ T B RdyQ
  /\ WF_WQ_WQSpec B Delta
  /\ exists TM : ThrdMem, exists Mcth : Mem, MemSplit TM (S_Mem S)
    /\ ThrdMemMod.maps_to TM tid Mcth
    /\ WF_Curr_Thrd Psi (S_Mem_Upd S Mcth) langid theta T tid
    /\ WF_ReadyQ Psi T RdyQ TM
    /\ WF_WQSet Psi T B TM.


Inductive embed_WQSet (T : TQ) : WQSet -> Mem -> Prop :=
    wq_emp  : embed_WQSet T WQSetMod.emp MemMod.emp
  | wq_cons : forall B l Q M M' M'',
                  embed_WQSet T B M 
                  -> WQSetMod.not_in_dom l B
                  -> ptoTQ l (filterTQ T Q) M'
                  -> MemMod.disjUnion M M' M''
                  -> embed_WQSet T (WQSetMod.put B l Q) M''.

Definition embed_RdyQ (T : TQ) (RdyQ : WQ) : MPred :=
    ptoTQ ready_Q (filterTQ T RdyQ).

(* curr_thrd --> tid *)
(* tid --> prev, next, tcb *)
Definition embed_curr_TCB (T : TQ) (tid : TID) : MPred :=
  fun (M : Mem) =>
    exists tcb,
      TQMod.maps_to T tid tcb
      /\ (mstar (ptoW curr_thrd tid) 
                (mstar (mstar (ptoxW tid) (ptoxW (tid+2))) (ptoTCB (tid+4) tcb))
         ) M.

Definition Itp_AIM (langid : LangID) (theta : Spec) (Psi : PSpec) (S : State) : Prop :=
  exists T, exists B, exists RdyQ, exists tid,
    pmstar (fun S => WF_AIM_World Psi S T B RdyQ tid langid theta) 
           (mstar (embed_curr_TCB T tid) 
                  (mstar (embed_RdyQ T RdyQ) 
                         (embed_WQSet T B)
                  )
           )
    S.

(* by guoyu, related lemmas are in Linking_Lemmas3.v *)
Definition embed_BlkQ (l : Label) (T : TQ) (BlkQ : WQ) : MPred :=
  ptoTQ l (filterTQ T BlkQ).

(* by guoyu, related lemmas are in Linking_Lemmas3.v *)
Definition embed_TCB (tid : TID) (tcb : Thread) : MPred :=
     (      tid |->w_     ) *^*          (* prev_ptr *)
     (  tid + 2 |->w_     ) *^*          (* next_ptr *)
     (ptoTCB (tid+4) tcb).


(* precondition for switch *)
(* stack space: 4 words, not including pc *)
(* here we need WF_TQ to know that all tid's are greater than 0 *)

(* note we do not need to specify pc in function specifications *)

Definition SW_MPre (T : TQ) (RdyQ : WQ) (tid : TID) (sp : Label) (pc : PC) (junk : int) :=
  (embed_curr_TCB T tid) *^* (embed_RdyQ T RdyQ) *^* (sp |-> pc) *^* (sp + 1 |-> junk) *^* (stkSpacePrec sp 8).

Definition SW_Prd_aux (T : TQ) (RdyQ : WQ) (tid : TID) (sp : Label) (tcb : Thread) (pc : PC) (junk : int) :=
  fun (S : State) =>
      (SW_MPre T RdyQ tid sp pc junk) (S_Mem S)
      /\ (TQMod.maps_to T tid tcb)
      /\ (subsetQ_TQ RdyQ T)
      /\ (S_RF S SP = sp) 
      /\ (WQMod.not_in_dom tid RdyQ) 
      /\ (WF_TQ T). 

Definition SW_Prd (S : State) :=
    exists T, exists RdyQ, exists tid, exists sp, exists tcb, exists pc, 
      (exists junk, ((SW_MPre T RdyQ tid sp pc junk)*^*mTrue) (S_Mem S))
      /\ (TQMod.maps_to T tid tcb)
      /\ (subsetQ_TQ RdyQ T)
      /\ (S_RF S SP = sp) 
      /\ (WQMod.not_in_dom tid RdyQ) 
      /\ (WF_TQ T).


Definition SW_MPost (T : TQ) (RdyQ : WQ) (tid : TID) (sp0: Label) (pc0 : PC) (junk : int) :=
  (embed_curr_TCB T tid) *^* (embed_RdyQ T RdyQ) *^* (sp0 |-> pc0) *^* (sp0 + 1 |-> junk) *^* (stkSpacePrec sp0 8).

Definition SW_Post_aux (T: TQ) (RdyQ : WQ) (tid : TID) (sp sp0 : Label) (tcb : Thread) (pc0 : PC) (junk : int) :=
  fun (S : State) =>
    (SW_MPost T RdyQ tid sp0 pc0 junk) (S_Mem S)
    /\ TQMod.maps_to T tid tcb
    /\ TCB_SP tcb = sp
    /\ S_RF S SP = sp
    /\ WQMod.not_in_dom tid RdyQ.

Definition SW_grt (S S' : State) :=
    forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
      -> forall T T' RdyQ tid sp tcb pc junk,
           SW_Prd_aux T RdyQ tid sp tcb pc junk (S_Mem_Upd S M0)
           -> T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
           -> exists M0', exists RdyQ', exists tid', exists sp', exists tcb',
                 MemMod.disjUnion M0' M1 (S_Mem S')
                 /\ addThrd RdyQ tid = addThrd RdyQ' tid'
                 /\ SW_Post_aux T' RdyQ' tid' sp' sp tcb' pc junk (S_Mem_Upd S' M0').

Definition SWITCH_Spec := Ast SW_Prd SW_grt 0.

(* block function requires 10 bytes of stack space *)
(* not including its own return addr. and args     *)
Definition BLK_MPre (T : TQ) (RdyQ BlkQ: WQ) (tid : TID) (sp l : Label) (pc : PC) (junk1 junk2 : int) :=
  (embed_curr_TCB T tid) *^* (embed_RdyQ T RdyQ) *^* (ptoTQ l (filterTQ T BlkQ))
  *^* (sp |-> pc) *^* (sp + 1 |-> junk1) *^* (sp + 2 |-> l) *^* (sp + 3 |-> junk2) *^* (stkSpacePrec sp 10).

Definition BLK_MPost (T : TQ) (RdyQ BlkQ : WQ) (tid : TID) (sp0 l : Label) (pc0 : PC) (junk1 junk2 : int) :=
  (embed_curr_TCB T tid) *^* (embed_RdyQ T RdyQ) *^* (ptoTQ l (filterTQ T BlkQ))
  *^* (sp0 |-> pc0) *^* (sp0 + 1 |-> junk1) *^* (sp0 + 2 |-> l) *^* (sp0 + 3 |-> junk2) *^* (stkSpacePrec sp0 10).

Definition BLK_Prd_aux (T : TQ) (RdyQ BlkQ: WQ) (tid : TID) (sp l : Label) (tcb : Thread) (pc : PC) (junk1 junk2 : int) :=
  fun (S : State) =>
      (BLK_MPre T RdyQ BlkQ tid sp l pc junk1 junk2) (S_Mem S)
      /\ exists_Idle T RdyQ tid
      /\ TQMod.maps_to T tid tcb
      /\ subsetQ_TQ RdyQ T
      /\ subsetQ_TQ BlkQ T
      /\ TCB_Status tcb = READY
      /\ S_RF S SP = sp
      /\ WQMod.not_in_dom tid RdyQ
      /\ WQMod.not_in_dom tid BlkQ
      /\ WQMod.disjoint RdyQ BlkQ
      /\ WF_TQ T.
      (* /\ S_Flag S _IF = false.*)
    
Definition BLK_Prd (S : State) :=
    exists T, exists RdyQ, exists BlkQ, exists tid, exists sp, exists l, exists tcb, exists pc, 
      (exists junk1, exists junk2,
       ((BLK_MPre T RdyQ BlkQ tid sp l pc junk1 junk2) *^* mTrue) (S_Mem S))
      /\ exists_Idle T RdyQ tid
      /\ TQMod.maps_to T tid tcb
      /\ subsetQ_TQ RdyQ T
      /\ subsetQ_TQ BlkQ T
      /\ TCB_Status tcb = READY
      /\ S_RF S SP = sp
      /\ WQMod.not_in_dom tid RdyQ
      /\ WQMod.not_in_dom tid BlkQ
      /\ WQMod.disjoint RdyQ BlkQ
      /\ WF_TQ T.

(*
Definition BLK_Prd (S : State) :=
    exists T, exists RdyQ, exists BlkQ, exists tid, exists sp, exists l, exists tcb, exists pc,
      imprec (BLK_Prd_aux T RdyQ BlkQ tid sp l tcb pc) S.
*)

Definition BLK_Post_aux (T : TQ) (RdyQ BlkQ : WQ) (tid : TID) (sp sp0 l : Label) (tcb : Thread) (pc0 : PC) (junk1 junk2 : int) :=
  fun (S : State) =>
     TQMod.maps_to T tid tcb
  /\ TCB_SP tcb = sp
  /\ S_RF S SP = sp
  /\ WQMod.not_in_dom tid RdyQ
  (* /\ S_Flag S _IF = false *)
  /\ (BLK_MPost T RdyQ BlkQ tid sp0 l pc0 junk1 junk2) (S_Mem S).

Definition BLK_grt (S S' : State) :=
    forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
      -> forall T T' RdyQ BlkQ BlkQ' tid sp l tcb pc junk1 junk2,
           BLK_Prd_aux T RdyQ BlkQ tid sp l tcb pc junk1 junk2 (S_Mem_Upd S M0)
           -> T' = TQMod.put T tid (sp, BLOCKED, pc)
           -> BlkQ' = addThrd BlkQ tid
           -> exists M0', exists RdyQ', exists tid', exists sp', exists tcb',
                 MemMod.disjUnion M0' M1 (S_Mem S')
                 /\ RdyQ = addThrd RdyQ' tid'
                 /\ (BLK_Post_aux T' RdyQ' BlkQ' tid' sp' sp l tcb' pc junk1 junk2) (S_Mem_Upd S' M0').

Definition BLK_Spec := Ast BLK_Prd BLK_grt 0.

(* block function requires 10 bytes of stack space *)
(* not including its own return addr. and args     *)
Definition UNBLK_MPre (T : TQ) (RdyQ BlkQ: WQ) (sp l : Label) (junk : int) :=
  (embed_RdyQ T RdyQ) *^* (ptoTQ l (filterTQ T BlkQ)) *^* (sp + 2 |-> l) *^* (sp + 3 |-> junk) *^* (stkSpacePrec sp 10).

Definition UNBLK_MPost := UNBLK_MPre.

Definition UNBLK_Prd_aux (T : TQ) (RdyQ BlkQ: WQ) (sp l : Label) (junk : int) :=
  fun (S : State) =>
      (UNBLK_MPre T RdyQ BlkQ sp l junk) (S_Mem S)
      /\ subsetQ_TQ RdyQ T
      /\ subsetQ_TQ BlkQ T
      /\ S_RF S SP = sp
      /\ WQMod.disjoint RdyQ BlkQ
      /\ WF_TQ T.
      (* /\ S_Flag S _IF = false.*)


Definition UNBLK_Prd (S : State) :=
    exists T, exists RdyQ, exists BlkQ, exists sp, exists l, 
      (exists junk,
       ((UNBLK_MPre T RdyQ BlkQ sp l junk) *^* mTrue) (S_Mem S))
      /\ subsetQ_TQ RdyQ T
      /\ subsetQ_TQ BlkQ T
      /\ S_RF S SP = sp
      /\ WQMod.disjoint RdyQ BlkQ
      /\ WF_TQ T.


Definition UNBLK_Post_aux (T : TQ) (RdyQ0 BlkQ0 : WQ) (sp bp l : Label) (junk : int) :=
  fun (S : State) => 
    S_RF S SP = sp /\ S_RF S BP = bp (* /\ S_Flag S _IF = false*)
    /\ (filterTQ T BlkQ0 = TQMod.emp /\ UNBLK_MPost T RdyQ0 BlkQ0 sp l junk (S_Mem S) /\ S_RF S AX = 0
        \/ exists tid, exists tcb,
                       WQMod.in_dom tid BlkQ0
                       /\ T tid = Some tcb
                       /\ UNBLK_MPost (TQMod.put T tid (TCB_SP tcb, READY, TCB_PC tcb)) (addThrd RdyQ0 tid) (WQMod.remove BlkQ0 tid) sp l junk (S_Mem S)
                       /\ S_RF S AX <> 0).

(*
Definition UNBLK_Post_aux (T : TQ) (RdyQ0 BlkQ0 : WQ) (sp bp l : Label) (junk : int) :=
  fun (S : State) => 
    S_RF S SP = sp /\ S_RF S BP = bp (* /\ S_Flag S _IF = false*)
    /\ (BlkQ0 = WQMod.emp /\ UNBLK_MPost T RdyQ0 BlkQ0 sp l junk (S_Mem S) /\ S_RF S AX = 0
        \/ exists tid, exists tcb,
                       WQMod.in_dom tid BlkQ0
                       /\ T tid = Some tcb
                       /\ UNBLK_MPost (TQMod.put T tid (TCB_SP tcb, READY, TCB_PC tcb)) (addThrd RdyQ0 tid) (WQMod.remove BlkQ0 tid) sp l junk (S_Mem S)
                       /\ S_RF S AX <> 0).
*)

Definition UNBLK_grt (S S' : State) :=
    forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
      -> forall T RdyQ BlkQ sp l junk,
           UNBLK_Prd_aux T RdyQ BlkQ sp l junk (S_Mem_Upd S M0)
           -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') /\
                 UNBLK_Post_aux T RdyQ BlkQ sp (S_RF S BP) l junk (S_Mem_Upd S' M0').


Definition UNBLK_Spec := Ast UNBLK_Prd UNBLK_grt 0.


Definition SWITCH_Spec0 := packS0 rho_prim SWITCH_Spec.
Definition BLK_Spec0    := packS0 rho_prim BLK_Spec.
Definition UNBLK_Spec0  := packS0 rho_prim UNBLK_Spec.

Definition ext_Prim_Specs (Psi : PSpec) : PSpec :=
  consCT _UNBLK UNBLK_Spec0 (consCT _BLOCK BLK_Spec0 (consCT _SWITCH SWITCH_Spec0 Psi)).

Definition Minimal_D (D : LDict) : Prop :=
  D rho_ready = SomeT1 _ (Itp_AIM rho_ready) 
  /\ D rho_idle = SomeT1 _ (Itp_AIM rho_idle) 
  /\ D rho_ih = SomeT1 _ (Itp_AIM rho_ih).

Definition ext_IH_Spec (Psi : PSpec) : PSpec :=
  (consCT h_entry IH_Spec0 Psi).

Definition ext_Spec (Psi : PSpec) (f : flab) (theta : Spec0) : PSpec :=
  consCT f theta Psi.

Parameter SW_BLK_neq : _SWITCH <> _BLOCK.
Parameter SW_UNBLK_neq : _SWITCH <> _UNBLK.
Parameter BLK_UNBLK_neq : _BLOCK <> _UNBLK.
Hint Resolve SW_BLK_neq SW_UNBLK_neq BLK_UNBLK_neq.
